""" Low-level process information APIs

    (c) Copyright Marc-Andre Lemburg; All Rights Reserved.
    See the documentation for further information on copyrights,
    or contact the author (mal@lemburg.com).

"""
import os,string

def procstat(pid=None,

	     int=int,getpid=os.getpid,split=string.split):

    """ Direct interface to /proc/<pid>/stat available on Linux.
    
        Whenever the format changes, this function will fail. It is
        also NOT portable; but better than nothing, since
        resource.getrusage() doesn't return meaningful data on Linux.

        Returns a tuple consisting of the following entries (all
        numbers appear as numbers, strings as strings). The following
        information is an edited version of part of the Linux man-page
        for proc(5):

	Position:     Field:

        0             pid %d The process id.

        1             comm %s
                             The filename of the executable, in
                             parentheses.  This is visible
                             whether or not the executable is
                             swapped out.

        2             state %c
                             One character from the string
                             "RSDZT" where R is running, S is
                             sleeping in an interruptible wait, D
                             is sleeping in an uninterruptible
                             wait or swapping, Z is zombie, and T
                             is traced or stopped (on a signal).

        3             ppid %d
                             The PID of the parent.

        4             pgrp %d
                             The process group ID of the process.

        5             session %d
                             The session ID of the process.

        6             tty %d The tty the process uses.

        7             tpgid %d
                             The process group ID of the process
                             which currently owns the tty that
                             the process is connected to.

        8             flags %u
                             The flags of the process.
                             Currently, every flag has the math
                             bit set, because crt0.s checks for
                             math emulation, so this is not
                             included in the output.  This is
                             probably a bug, as not every process
                             is a compiled C program.  The math
                             bit should be a decimal 4, and the
                             traced bit is decimal 10.

        9             minflt %u
                             The number of minor faults the
                             process has made, those which have
                             not required loading a memory page
                             from disk.

        10            cminflt %u
                             The number of minor faults that the
                             process and its children have made.

        11            majflt %u
                             The number of major faults the
                             process has made, those which have
                             required loading a memory page from
                             disk.
	
        12            cmajflt %u
                             The number of major faults that the
                             process and its children have made.

        13            utime %d
                             The number of jiffies that this
                             process has been scheduled in user
                             mode.

        14            stime %d
                             The number of jiffies that this
                             process has been scheduled in kernel
                             mode.

        15            cutime %d
                             The number of jiffies that this
                             process and its children have been
                             scheduled in user mode.

        16            cstime %d
                             The number of jiffies that this
                             process and its children have been
                             scheduled in kernel mode.

        17            counter %d
                             The current maximum size in jiffies
                             of the process's next timeslice, or
                             what is currently left of its
                             current timeslice, if it is the
                             currently running process.

        18            priority %d
                             The standard nice value, plus
                             fifteen.  The value is never
                             negative in the kernel.

        19            timeout %u
                             The time in jiffies of the process's
                             next timeout.

        20            itrealvalue %u
                             The time (in jiffies) before the
                             next SIGALRM is sent to the process
                             due to an interval timer.

        21            starttime %d Time the process started in
                             jiffies after system
                             boot.

        22            vsize %u
                             Virtual memory size

        23            rss %u Resident Set Size: number of pages
                             the process has in real memory,
                             minus 3 for administrative purposes.
                             This is just the pages which count
                             towards text, data, or stack space.
                             This does not include pages which
                             have not been demand-loaded in, or
                             which are swapped out.

        24            rlim %u
                             Current limit in bytes on the rss of
                             the process (usually 2,147,483,647).

        25            startcode %u
                             The address above which program text
                             can run.

        26            endcode %u
                             The address below which program text
                             can run.

        27            startstack %u
                             The address of the start of the
                             stack.

        28            kstkesp %u
                             The current value of esp (32-bit
                             stack pointer), as found in the
                             kernel stack page for the process.

        29            kstkeip %u
                             The current EIP (32-bit instruction
                             pointer).

        30            signal %d
                             The bitmap of pending signals
                             (usually 0).

        31            blocked %d
                             The bitmap of blocked signals
                             (usually 0, 2 for shells).

        32            sigignore %d
                             The bitmap of ignored signals.

        33            sigcatch %d
                             The bitmap of catched signals.

        34            wchan %u
                             This is the "channel" in which the
                             process is waiting.  This is the
                             address of a system call, and can be
                             looked up in a namelist if you need
                             a textual name.  (If you have an up-
                             to-date /etc/psdatabase, then try ps
                             -l to see the WCHAN field in action)

        These positions are exported by the module as constants.  They
        all have the prefix PS and are in named according to the above
        listing but using uppercase letters, e.g. the position of
        priority is stored in PSPRIORITY.

    """
    if pid is None:
	pid = getpid()
    stat = split(open('/proc/%i/stat' % pid).read())
    for i in range(len(stat)):
	try:
	    stat[i] = int(stat[i])
	except ValueError:
	    pass
    return stat

### Constants for the procstat sequence

procstat_consts = ['PSPID', 'PSCOMM', 'PSSTATE', 'PSPPID', 'PSPGRP',
'PSSESSION', 'PSTTY', 'PSTPGID', 'PSFLAGS', 'PSMINFLT', 'PSCMINFLT',
'PSMAJFLT', 'PSCMAJFLT', 'PSUTIME', 'PSSTIME', 'PSCUTIME', 'PSCSTIME',
'PSCOUNTER', 'PSPRIORITY', 'PSTIMEOUT', 'PSITREALVALUE',
'PSSTARTTIME', 'PSVSIZE', 'PSRSS', 'PSRLIM', 'PSSTARTCODE',
'PSENDCODE', 'PSSTARTSTACK', 'PSKSTKESP', 'PSKSTKEIP', 'PSSIGNAL',
'PSBLOCKED', 'PSSIGIGNORE', 'PSSIGCATCH', 'PSWCHAN']

def _init_consts():
    g = globals()
    for i in range(len(procstat_consts)):
	const = procstat_consts[i]
	g[const] = i
_init_consts()

###

try:
    import resource
except ImportError:
    resource = None

def procsize():

    """ Returns the virtual process size in bytes or 0 in case it
        cannot determine the value.

	This function first tries to use the resource module and if
	that fails reverts to using procstat().

    """
    if resource:
	usage = resource.getrusage(resource.RUSAGE_SELF)
	size = usage[3] + usage[4] + usage[5]
    else:
	size = 0
    if size == 0:
	try:
	    size = procstat()[22]
	except IOError:
	    size = 0
    return size

