#!/usr/bin/env python -O

""" Script to test database capabilities and the mxODBC interface
    for functionality and memory leaks.

    Synopsis:

    python test.py [<subpackage>] [-q]

    <subpackage> ... name of the subpackage to default to
    -q ... quiet operation, i.e. no error information is displayed

    To enable mxODBC logging (provided mxODBC was compiled with
    debugging enabled; see docs):

    python -d test.py [-q]

    (c) Copyright Marc-Andre Lemburg; All Rights Reserved.
    See the documentation for further information on copyrights,
    or contact the author (mal@lemburg.com).
    
"""
import exceptions,string,sys,traceback,time
import DateTime
import ODBC,ODBC.Misc.proc

# Globals
verbosity = 2
id = 0

class Error(exceptions.StandardError):
    pass

class PrecisionError(Error):
    pass

class DatabaseTest:

    def __init__(self,name,database,connection,*args,**kws):

	self.name = name
	self.database = database
	self.connection = connection
	self.args = args
	self.kws = kws
	self.cursor = connection.cursor()

    def prepare(self,*args,**kws):

	""" Create tables needed for the test.

	    args and kws are the additional arguments ones passed to
	    the constructor.

	"""
	pass

    def test(self):

	""" Test implementation.
	"""
	pass

    def run(self):

	""" Run the test.
	"""
	try:
	    rc = apply(self.prepare,self.args,self.kws)
	    if rc:
		rc = self.test()
	finally:
	    self.cleanup()
	return rc

    def cleanup(self):

	""" Drop all tables possibly created by the test.
	"""
	pass

    def table_exists(self,name):

	try:
	    self.cursor.execute('select * from %s where 1=0' % name)
	except self.database.Error:
	    return 0
	else:
	    return 1
    
    def new_table_name(self):

	global id
	i = id + 1
	while 1:
	    name = 'mxODBC%04i' % i
	    if not self.table_exists(name):
		id = i
		return name
	    i = i + 1

class DataTest(DatabaseTest):

    rows = 10
    
    messages = ('supported',
		'type not supported',
		'unexpected database error',
		'not supported',
		'unexpected database warning',
		'precision lossage')

    def prepare(self,columndefs,generator,verifier=None,
		messages=None,verbose=None):

	""" Create a table using a list of column definitions given in
	    columndefs.
	
	    generator must be a function taking arguments (row_number,
	    col_number) returning a suitable data object for insertion
	    into the table.

	    verifier may be given as function taking three arguments
	    (row_number, col_number,data). It has to verify that data
	    is considered a correct output and return the result as
	    truth value. The function should take the corresponding
	    generator value for row_number and col_number into
	    account. If not given, data must compared equal to
	    generator(row_number,col_number).

	    If given, messages must be a sequence having the following
	    string entries which are printed to report different
	    situations:
	    
	    'Test ran ok.',
	    'Test preparation failed.',
	    'Test run failed due to database error.',
	    'Test run failed due to data mismatch.',
	    'Test run failed due to database warning',
	    'Test run failed due to precision lossage'

	"""
	self.table = self.new_table_name()
	self.columndefs = columndefs
	self.generator = generator

	if verifier is None:
	    self.verifier = lambda row,col,data,generator=generator: \
			    data == generator(row,col)
	else:
	    self.verifier = verifier

	if messages is not None:
	    if len(messages) < 6:
		raise TypeError,'need at least 4 message strings'
	    self.messages = messages

	if verbose is not None:
	    self.verbose = verbose
	else:
	    self.verbose = verbosity

	# Create the test table
	try:
	    self.cursor.execute('create table %s (%s)' % 
				(self.table,
				 string.join(columndefs,', ')))
	except (self.database.ProgrammingError,
		self.database.OperationalError):
	    print '%-45s: %s' % (self.name,self.messages[1])
	    return 0
	return 1

    def insert_data(self):

	insert_statement = ('insert into %s values (%s)' % 
			    (self.table,
			     string.join(['?'] * len(self.columndefs),', ')))
	for i in range(self.rows):
	    data = []
	    for j in range(len(self.columndefs)):
		data.append(self.generator(i,j))
	    self.cursor.execute(insert_statement,tuple(data))

    def verify_data(self):

	self.cursor.execute('select * from %s' % self.table)
	l = self.cursor.fetchall()
	for i in range(self.rows):
	    for j in range(len(self.columndefs)):
		origdata = self.generator(i,j)
		fetcheddata = l[i][j]
		try:
		    ok = self.verifier(i,j,fetcheddata)
		except TypeError:
		    raise Error(
			'column %i of row %i: '
			'input type %s, output type %s' %
			(j,i,
			 type(origdata).__name__,
			 type(fetcheddata).__name__))
		if not ok:
		    try:
			origlen = len(origdata)
			fetchedlen = len(fetcheddata)
		    except TypeError:
			origlen = '?'
			fetchedlen = '?'
		    try:
			diff = str(fetcheddata - origdata)
		    except (TypeError,ValueError):
			diff = '?'
		    why = ('column %i of row %i: '
			   'input data: %s (type=%s,len=%s), '
			   'output data: %s (type=%s,len=%s), '
			   'output - input: %s' %
			   (j,i,
			    repr(origdata),
			    type(origdata).__name__,
			    origlen,
			    repr(fetcheddata),
			    type(fetcheddata).__name__,
			    fetchedlen,
			    diff))
		    if type(origdata) == type(fetcheddata) and \
		       diff != '?':
			raise PrecisionError,why
		    else:
			raise Error,why

    def test(self):

	try:
	    self.insert_data()
	    self.verify_data()
	except self.database.DataError,why:
	    if self.verbose > 0:
		print '%-45s: %s [%s: %s]' % \
		      ((self.name,self.messages[3],) + sys.exc_info()[:2])
	    else:
		print '%-45s: %s' % (self.name,self.messages[3])
	    return 0
	except self.database.Error,why:
	    if self.verbose > 0:
		print '%-45s: %s [%s: %s]' % \
		      ((self.name,self.messages[2],) + sys.exc_info()[:2])
	    else:
		print '%-45s: %s' % (self.name,self.messages[2])
	    return 0
	except self.database.Warning,why:
	    if self.verbose > 0:
		print '%-45s: %s [%s: %s]' % \
		      ((self.name,self.messages[4],) + sys.exc_info()[:2])
	    else:
		print '%-45s: %s' % (self.name,self.messages[4])
	    return 0
	except PrecisionError,why:
	    if self.verbose > 1:
		print '%-45s: %s [%s]' % (self.name,self.messages[5],why)
	    else:
		print '%-45s: %s' % (self.name,self.messages[5])
	    return 0
	except Error,why:
	    if self.verbose > 1:
		print '%-45s: %s [%s]' % (self.name,self.messages[3],why)
	    else:
		print '%-45s: %s' % (self.name,self.messages[3])
	    return 0
	print '%-45s: %s' % (self.name,self.messages[0])
	return 1

    def cleanup(self):

	try:
	    self.cursor.execute('drop table %s' % (self.table))
	except self.database.Error:
	    pass

class ArrayDataTest(DataTest):

    def insert_data(self):

	insert_statement = ('insert into %s values (%s)' % 
			    (self.table,
			     string.join(['?'] * len(self.columndefs),', ')))
	# Prepare data
	data = []
	for i in range(self.rows):
	    row = []
	    for j in range(len(self.columndefs)):
		row.append(self.generator(i,j))
	    data.append(tuple(row))

	# Execute the INSERT
	self.cursor.execute(insert_statement,data)

### Tests

def testNULL(Database,connection):

    """ NULLable columns """

    def generator(row,col):
	return None
    test = DataTest(' NULLable columns',
		    Database,connection,
		    ('col1 char(255) null',),
		    generator)
    return test.run()

def testCHARACTER(Database,connection):

    """ CHAR column type """

    def generator(row,col):
	return ('%i' % ((row+col) % 10)) * 255
    test = DataTest(' CHAR column type',
		    Database,connection,
		    ('col1 char(255)',),
		    generator)
    return test.run()

def testBinaryCHARACTER(Database,connection):

    """ CHAR column type with binary data """

    def generator(row,col):
	return ('%c' % ((row+col) % 256)) * 255
    test = DataTest(' CHAR column type with binary data',
		    Database,connection,
		    ('col1 char(255)',),
		    generator,
		    verbose=1)
    return test.run()

def testCHARACTERPadding(Database,connection):

    """ CHAR is padded with spaces """

    def generator(row,col):
	return ('%i' % ((row+col) % 256)) * (row % 255)
    def verifier(row,col,data,generator=generator):
	return data == ('%-255s' % generator(row,col))
    test = DataTest(' CHAR padding (with spaces)',
		    Database,connection,
		    ('col1 char(255)',),
		    generator,
		    verifier)
    return test.run()

def testVARCHAR(Database,connection):

    """ VARCHAR column type """

    def generator(row,col):
	if col == 0:
	    return ('%i' % ((row+col) % 10)) * 1024
	else:
	    return '%i,%i' % (col,row)
    test = DataTest(' VARCHAR column type',
		    Database,connection,
		    ('col1 varchar(1024)','col2 varchar(255)'),
		    generator)
    return test.run()

def testBinaryVARCHAR(Database,connection):

    """ VARCHAR column type with binary data """

    def generator(row,col):
	if col == 0:
	    return ('%i\000' % ((row+col) % 10)) * 512
	else:
	    return '%i\000%i' % (col,row)
    test = DataTest(' VARCHAR column type with binary data',
		    Database,connection,
		    ('col1 varchar(1024)','col2 varchar(255)'),
		    generator,
		    verbose=1)
    return test.run()

def testVARBINARY(Database,connection):

    """ VARBINARY column type """

    def generator(row,col):
	if col == 0:
	    return ('%i\000' % ((row+col) % 10)) * 512
	else:
	    return '%i\000%i' % (col,row)
    test = DataTest(' VARBINARY column type',
		    Database,connection,
		    ('col1 varbinary(1024)','col2 varbinary(255)'),
		    generator,
		    verbose=1)
    return test.run()

def testBINARY(Database,connection):

    """ BINARY column type """

    def generator(row,col):
	if col == 0:
	    return ('%i\000' % ((row+col) % 10)) * 512
	else:
	    return '%i\000%i' % (col,row)
    test = DataTest(' BINARY column type',
		    Database,connection,
		    ('col1 binary(1024)','col2 binary(255)'),
		    generator,
		    verbose=1)
    return test.run()

def testDECIMAL(Database,connection):

    """ DECIMAL column type """

    # Number data
    def generator(row,col):
	if col == 0:
	    return row
	else:
	    return float(row) + 0.51

    test = DataTest(' DECIMAL column type',
		    Database,connection,
		    ('col1 decimal(5,0)','col2 decimal(5,2)'),
		    generator)
    return test.run()

def testDOUBLE(Database,connection):

    """ DOUBLE column type """

    # Number data
    def generator(row,col):
	if col == 0:
	    return row
	else:
	    return float(row) + 0.51

    test = DataTest(' DOUBLE column type',
		    Database,connection,
		    ('col1 double','col2 double'),
		    generator)
    return test.run()

def testFLOAT(Database,connection):

    """ FLOAT column type """

    # Number data
    def generator(row,col):
	if col == 0:
	    return row
	else:
	    return float(row) + 0.51

    test = DataTest(' FLOAT column type',
		    Database,connection,
		    ('col1 float','col2 float'),
		    generator)
    return test.run()

def testBIT(Database,connection):

    """ BIT column type """

    # Number data
    def generator(row,col):
	return row % 2

    test = DataTest(' BIT column type',
		    Database,connection,
		    ('col1 bit',),
		    generator)
    return test.run()

def testINTEGER(Database,connection):

    """ INT column type """

    # Number data
    def generator(row,col):
	if col == 0:
	    return row
	else:
	    return -row

    test = DataTest(' INT column type',
		    Database,connection,
		    ('col1 int','col2 integer'),
		    generator)
    return test.run()

def testArrayINTEGER(Database,connection):

    """ INT column type using array processing """

    # Number data
    def generator(row,col):
	if col == 0:
	    return row
	else:
	    return -row

    test = ArrayDataTest(' INT column type using array processing',
			 Database,connection,
			 ('col1 int','col2 integer'),
			 generator)
    return test.run()

def testBIGINT(Database,connection):

    """ BIGINT column type with 64bits """

    # Number data
    def generator(row,col):
	if col == 0:
	    return row * 2
	else:
	    if row % 1:
		return 2L**63 - row - 1
	    else:
		return - 2L**63 + row

    test = DataTest(' BIGINT column type with 64bits',
		    Database,connection,
		    ('col1 bigint','col2 bigint'),
		    generator)
    return test.run()

def testDATETIME(Database,connection):

    """ DATETIME column type """

    def generator(row,col):
	return DateTime.DateTime(1999+row,3,2,11,48,3)
    test = DataTest(' DATETIME column type',
		    Database,connection,
		    ('col1 datetime',),
		    generator)
    return test.run()

def testDATETIMEWithStrings(Database,connection):

    """ DATETIME column type using string values """

    def generator(row,col):
	return str(DateTime.DateTime(1999+row,3,2,11,48,3))
    test = DataTest(' DATETIME with string values',
		    Database,connection,
		    ('col1 datetime',),
		    generator)
    return test.run()

def testDATE(Database,connection):

    """ DATE column type """

    # Date, Time...
    def generator(row,col):
	return DateTime.Date(1999+row,1,2)
    test = DataTest(' DATE column type',
		    Database,connection,
		    ('col1 date',),
		    generator)
    return test.run()

def testTIME(Database,connection):

    """ TIME column type """

    # Date, Time...
    def generator(row,col):
	return DateTime.Time(12,(56+row) % 60,(13+row) % 60)
    test = DataTest(' TIME column type',
		    Database,connection,
		    ('col1 time',),
		    generator)
    return test.run()

def testTIMESTAMP(Database,connection):

    """ TIMESTAMP column type """

    # Timestamps without fraction...
    def generator(row,col):
	return DateTime.Timestamp(1999+row,3,2,11,48,3)
    test = DataTest(' TIMESTAMP column type',
		    Database,connection,
		    ('col1 timestamp',),
		    generator)
    return test.run()

def testTIMESTAMPWithFraction(Database,connection):

    """ TIMESTAMP data having fractions """

    # Timestamps with fraction...
    def generator(row,col):
	return DateTime.Timestamp(1999+row,3,2,11,48,3.5)
    test = DataTest(' TIMESTAMP data having fractions',
		    Database,connection,
		    ('col1 timestamp',),
		    generator)
    return test.run()

def testIMAGE(Database,connection):

    """ IMAGE column type """

    def generator(row,col):
	return 'IMAGE Data \000\001' * 1024
    test = DataTest(' IMAGE column type',
		    Database,connection,
		    ('col1 image',),
		    generator,
		    verbose=1)
    return test.run()

def testLONG(Database,connection):

    """ LONG column type """

    # Knows LONGs...
    def generator(row,col):
	return 'BLOB Text ' * 1024
    test = DataTest(' LONG column type',
		    Database,connection,
		    ('col1 long',),
		    generator,
		    verbose=1)
    return test.run()

def testLargeLONG(Database,connection):

    """ LONGs with >32kB text data """

    def generator(row,col):
	return 'BLOB Text BLOB Text BLOB Text BLOB Text ' * 1024
    test = DataTest(' LONGs with >32kB text data',
		    Database,connection,
		    ('col1 long',),
		    generator,
		    verbose=1)
    return test.run()
    
def testBinaryLONG(Database,connection):

    """ LONGs with binary data """

    def generator(row,col):
	return 'BLOB\000Binary ' * 1024
    test = DataTest(' LONGs with binary data',
		    Database,connection,
		    ('col1 long',),
		    generator,
		    verbose=1)
    return test.run()
    
def testLONGTEXT(Database,connection):

    """ LONGTEXT column type """

    # Knows LONGTEXTs...
    def generator(row,col):
	return 'BLOB Text ' * 1024
    test = DataTest(' LONGTEXT column type (text data)',
		    Database,connection,
		    ('col1 longtext',),
		    generator,
		    verbose=1)
    return test.run()

def testLONGBINARY(Database,connection):

    """ LONGBINARY column type """

    # Knows LONGs...
    def generator(row,col):
	return 'BLOB\000Binary ' * 1024
    test = DataTest(' LONGBINARY column type (binary data)',
		    Database,connection,
		    ('col1 longbinary',),
		    generator,
		    verbose=1)
    return test.run()

def testMEMO(Database,connection):

    """ MEMO column type """

    # Knows MEMOs...
    def generator(row,col):
	return 'BLOB Text ' * 1024
    test = DataTest(' MEMO column type',
		    Database,connection,
		    ('col1 memo',),
		    generator,
		    verbose=1)
    return test.run()

def testLargeMEMO(Database,connection):

    """ MEMOs with >32kB text data """

    def generator(row,col):
	return 'BLOB Text BLOB Text BLOB Text BLOB Text ' * 1024
    test = DataTest(' MEMOs with >32kB text data',
		    Database,connection,
		    ('col1 memo',),
		    generator,
		    verbose=1)
    return test.run()
    
def testBinaryMEMO(Database,connection):

    """ MEMOs with binary data """

    def generator(row,col):
	return 'BLOB\000Binary ' * 1024
    test = DataTest(' MEMOs with binary data',
		    Database,connection,
		    ('col1 memo',),
		    generator,
		    verbose=1)
    return test.run()
    
def testTEXT(Database,connection):

    """ TEXT column type """

    def generator(row,col):
	return 'BLOB Text ' * 1024
    test = DataTest(' TEXT column type',
		    Database,connection,
		    ('col1 text',),
		    generator,
		    verbose=1)
    return test.run()

def testLONG_BYTE(Database,connection):

    """ LONG BYTE column type """

    def generator(row,col):
	return 'BLOB\000Binary ' * 1024
    test = DataTest(' LONG BYTE column type',
		    Database,connection,
		    ('col1 long byte',),
		    generator,
		    verbose=1)
    return test.run()
    
def testLargeLONG_BYTE(Database,connection):

    """ LONG BYTE with >32kB text data """

    def generator(row,col):
	return 'BLOB Text BLOB Text BLOB Text BLOB Text ' * 1024
    test = DataTest(' LONG BYTE with >32kB text data',
		    Database,connection,
		    ('col1 long byte',),
		    generator,
		    verbose=1)
    return test.run()
    
def testBLOB(Database,connection):

    """ BLOB column type with binary data """

    def generator(row,col):
	return 'BLOB\000Binary ' * 1024
    test = DataTest(' BLOB column type with binary data',
		    Database,connection,
		    ('col1 blob',),
		    generator,
		    verbose=1)
    return test.run()

def testLargeBLOB(Database,connection):

    """ BLOBs with >32kB text data """

    def generator(row,col):
	return 'BLOB Text BLOB Text BLOB Text BLOB Text ' * 1024
    test = DataTest(' BLOBs with >32kB text data',
		    Database,connection,
		    ('col1 blob',),
		    generator,
		    verbose=1)
    return test.run()

def testTranscations(Database,connection):

    """ Transactions """

    print '%-45s:' % ' Transactions',
    try:
	connection.commit()
	# XXX Do something and check if the rollback really works.
	connection.rollback()
	print 'supported'
	return 1
    except AttributeError:
	print 'not supported'
	return 0
    except Database.NotSupportedError:
	print 'not supported'
	return 0
    except Database.Error:
	print 'unexpected error [%s: %s]' % sys.exc_info()[:2]
	return 0

def testBindMethod(Database,connection):

    """ Variable Bind Method """

    print '%-45s:' % ' Variable Bind Method',
    try:
	if connection.bindmethod == Database.BIND_USING_PYTHONTYPE:
	    print 'Python Type'
	elif connection.bindmethod == Database.BIND_USING_SQLTYPE:
	    print 'SQL Type'
	else:
	    print 'Unkown: %i' % connection.bindmethod
	return 1
    except AttributeError:
	print 'not supported'
	return 0
    except Database.OperationalError:
	print 'not supported'
	return 0
    except Database.Error:
	print 'unexpected error [%s: %s]' % sys.exc_info()[:2]
	return 0

###

def get_tests():

    tests = []
    for key,value in globals().items():
	if key[:4] == 'test':
	    if hasattr(value,'__doc__'):
		name = value.__doc__
	    else:
		try:
		    name = test.func_name
		except AttributeError:
		    name = test.im_func.func_name
		if name[:4] == 'test':
		    name = name[4:]
	    tests.append((name,value))
    tests.sort()
    return tests

def run_suite(Database,connection,tests=None):
    
    # Default for tests: all defined functions with 'test' prefix
    if tests is None:
        tests = get_tests()
    # Run tests
    for name,test in tests:
	test(Database,connection)

### Type Info

typeinfo_headers = [
    "TYPE_NAME", "DATA_TYPE", "COLUMN_SIZE", "LITERAL_PREFIX",
    "LITERAL_SUFFIX", "CREATE_PARAMS", "NULLABLE", "CASE_SENSITIVE",
    "SEARCHABLE", "UNSIGNED_ATTRIBUTE", "FIXED_PREC_SCALE",
    "AUTO_UNIQUE_VALUE", "LOCAL_TYPE_NAME", "MINIMUM_SCALE",
    "MAXIMUM_SCALE", "SQL_DATA_TYPE", "SQL_DATETIME_SUB",
    "NUM_PREC_RADIX", "INTERVAL_PRECISION", "USERTYPE"]

def print_typeinfo(Database,connection,sqltypecode=None):

    if sqltypecode is None:
	sqltypecode=Database.SQL.ALL_TYPES

    print 'Column Types Supported by the ODBC Driver/Manager:'
    print
    cursor = connection.cursor()
    if not hasattr(cursor,'gettypeinfo'):
	print '  No information available.'
	print
	return
    cursor.gettypeinfo(sqltypecode)
    ODBC.print_resultset(cursor,typeinfo_headers)

###

def main():

    print '='*72
    print 'mxODBC Test Suite'
    print '-'*72
    print

    if sys.platform == 'win32':
	pkg_default = 'Windows'
    elif len(sys.argv) > 1:
	pkg_default = sys.argv[1]
    else:
	pkg_default = 'MySQL'

    pkgname = raw_input('Subpackage Name [%s]: ' % pkg_default) or \
	      pkg_default
    try:
	Database = __import__('ODBC.%s' % pkgname,locals(),globals(),'*')
    except ImportError,why:
	print
	print 'Import of ODBC.%s failed: %s' % (pkgname,why)
	return 1

    if pkgname == 'Adabas':
	dc_default = 'DSN=degas:MYDB;UID=lemburg;PWD=lemburg'
	c_default = 'degas:MYDB,lemburg,lemburg'
	clearac_default = '1'
    elif pkgname == 'MySQL':
	dc_default = 'DSN=test;UID=root'
	c_default = 'test,root,'
	clearac_default = '0'
    elif pkgname == 'iODBC':
	dc_default = 'DSN=degas:MYDB;UID=lemburg;PWD=lemburg'
	c_default = 'adabas,lemburg,lemburg'
	clearac_default = '1'
    elif pkgname == 'Windows':
	dc_default = 'DSN=test;UID=test;PWD=test'
	c_default = 'test,test,test'
	clearac_default = '1'
    else:
	dc_default = 'DSN=datasource_name;UID=user_id;PWD=password'
	c_default = 'datasource_name,user_id,password'
	clearac_default = '1'

    if not hasattr(Database,'DriverConnect'):
	connectargs = raw_input(
	    'Connect arguments [%s]: ' % c_default) or \
	    c_default
	connectargs = tuple(map(string.strip,string.split(connectargs,',')))
	connectapi = Database.Connect
    else:
	connectargs = raw_input(
	    'DriverConnect arguments [%s]: ' % dc_default) or \
	    dc_default
	connectargs = (string.strip(connectargs),)
	connectapi = Database.DriverConnect

    clearac = raw_input('Clear AUTOCOMMIT  ? (1/0) [%s] ' % clearac_default) or \
	      clearac_default
    if clearac == '1':
	connectargs = connectargs + (1,)
    else:
	connectargs = connectargs + (0,)

    checkleaks = raw_input('Run tests continuously to check for leaks ? (y/n) [n] ')
    if checkleaks in ('n','N'):
	checkleaks = ''

    if checkleaks:
	reconnect = raw_input('Reconnect for every run ? (y/n) [n] ')
	if reconnect in ('n','N'):
	    reconnect = ''
    else:
	reconnect = 'y'

    typeinfo = raw_input('Show driver type information ? (y/n) [n] ')
    if typeinfo in ('n','N'):
	typeinfo = ''

    outputfile = raw_input('Output file [stdout]: ')
    if outputfile != '':
	sys.stdout = open(outputfile,'w')

    tests = get_tests()
    if checkleaks:
	print
	print 'Tests to perform:'
	remove = []
	for i in range(len(tests)):
	    name,test = tests[i]
	    ok = raw_input(' %s ? (y/n) [y] ' % name)
	    if ok in ('n','N'):
		remove.append(i)
	remove.reverse()
	for i in remove:
	    del tests[i]
    print

    connection = None
    run = 0
    vss_after_first_run = 0
    while 1:
	run = run + 1
	if checkleaks:
	    print 'Test suite run %i:' % run
	else:
	    print 'Test suite:'
	if not connection:
	    print ' Connecting to the database.'
	    connection = apply(connectapi,connectargs)
	    print ' Connected to DBMS: %s; Version: %s' % \
		  (connection.getinfo(Database.SQL.DBMS_NAME)[1],
		   connection.getinfo(Database.SQL.DBMS_VER)[1]
		   )
	    print '              ODBC driver: %s %s; ODBC Version: %s' % \
		  (connection.getinfo(Database.SQL.DRIVER_NAME)[1],
		   connection.getinfo(Database.SQL.DRIVER_VER)[1],
		   connection.getinfo(Database.SQL.DRIVER_ODBC_VER)[1],
		   )
	    print
	run_suite(Database,connection,tests)
	if reconnect:
	    print
	    print ' Disconnecting.'
	    connection = None
	if not checkleaks:
	    break
	print
	if 0 and pkgname == 'Adabas':
	    print 'Pause for memory leak detection...'
	    # Sleep a bit to let Adabas AUTOSAVE LOG clear the LOG
	    time.sleep(10)
	vss = ODBC.Misc.proc.procsize()
	if vss:
	    if not vss_after_first_run:
		vss_after_first_run = vss
	    leakage = vss - vss_after_first_run
	    print 'Current virtual process size:'
	    print '  %i bytes' % vss
	    print '  = First run %+i bytes' % leakage
	    print '  = Leakage per run: %i bytes' % (leakage / run)
	    print '  = %5.2f %% of the initial size' % \
		  ((100.0 * leakage) / vss_after_first_run )
	print

    # Print type infos
    if typeinfo:
	connection = apply(connectapi,connectargs)
	print
	print_typeinfo(Database,connection)
	connection = None

    return 0

if __name__ == '__main__':
    if '-q' in sys.argv:
	verbosity = 0
	sys.argv.remove('-q')
    try:
	rc = main()
    except KeyboardInterrupt:
	print
	import traceback
	traceback.print_exc()
	print
	print '* User break... terminating.'
	rc = 2
    time.sleep(1)
    sys.exit(rc)
