/* ---------------------------------------------------------------------
   mxODBC -- A generic ODBC 2.0 interface

   (c) Marc-Andre Lemburg; All Rights Reserved; mailto:mal@lemburg.com;
   See the documentation for further copyright information or contact
   the author.
   ---------------------------------------------------------------------
*/

/* Define this to aid in finding memory leaks */
/*#define MAL_MEM_DEBUG*/
/*#define MAL_DEBUG*/

/* Logging file used by debugging facility */
#ifndef MAL_DEBUG_OUTPUTFILE
# define MAL_DEBUG_OUTPUTFILE "mxODBC.log"
#endif

/* We want all our symbols to be exported */
#define MX_BUILDING_MXODBC

#include "mx.h"
#include "mxODBC.h"
#include "mxDateTime.h"
#include "mxSQLCodes.h"
#include <ctype.h>

/* XXX Don't forget to update the version number in dbi.py ! */
#define VERSION "1.1.1"

/* ---------------------------------------------------------------------
   Set defaults for some of the customization switches mxODBC offers.
   Most of these are managed and explained in the mxODBC.h header
   file.
   ---------------------------------------------------------------------
*/

/* Default size used as start size for fetching LONG columns; they are
   fetched using SQLGetData and the data buffer grows dynamically. */
#define DEFAULT_LONG_SIZE 1024

/* Default size for fields the are to be fetched but for which the 
   ODBC driver does not return a valid size. */
#define DEFAULT_FIELD_SIZE 64

/* Defines how the module should handle fractions in
   timestamps. Fraction integers are converted to a seconds double
   using this constant. The ODBC docs say that a fraction is one
   nanosecond while the IBM SAA SQL docs from 1988 refer to
   microseconds. Default is: 1 fraction == 1 nanosecond (1e-9) or
   1,000,000,000.0 fractions make up one second. The value has to be a
   double. */
#ifdef USE_MICROSECOND_FRACTIONS
# define FRACTIONS_IN_SECOND ((double) 1e6)
#else
# define FRACTIONS_IN_SECOND ((double) 1e9)
#endif

/* DB API threadlevel constant

   mxODBC is itself programmed in a thread safe way, yet this does not
   imply the thread safety of the underlying ODBC driver. The DB API
   specifies a module global that gives a programmer information about
   the level of thread safety supported. It defaults to 1 for mxODBC
   meaning that threads may share the module, but not the connections.

*/
#ifndef MXODBC_THREADLEVEL
# define MXODBC_THREADLEVEL 1
#endif

/* DB API apilevel constant

   mxODBC was started with DB API 1.0 in mind, but gradually shifted
   towards a modified API. Many of those ideas were incorporated into
   Python DB API 2.0.
   
   The constant has to be a string. Currently "1.0" and "2.0" are
   allowed.

*/
#ifndef MXODBC_APILEVEL
# define MXODBC_APILEVEL "2.0"
#endif

/* Variable input binding option.

   mxODBC can bind variables using the information from the
   driver/manager or by looking at the passed in Python types.
   
   Default is to use the driver/manager information
   (USE_SQLTYPE_BINDING).  This only works if the SQLDescribeParam()
   API is available.

*/
#if !defined(USE_SQLTYPE_BINDING) && !defined(USE_PYTHONTYPE_BINDING)
# define USE_SQLTYPE_BINDING
#endif

#ifdef DONT_HAVE_SQLDescribeParam
# define USE_PYTHONTYPE_BINDING
# undef USE_SQLTYPE_BINDING
#endif

/* This name is used in repr() as identifier for the used interface */
#ifndef MXODBC_INTERFACENAME
# define "ODBC"
#endif

/* Some drivers don't define this value (even though its standard ODBC
   2.0), e.g. Solid is one that doesn't. Since we use it in the
   binding code below, we simply redefine it to its standard value
   taken from WINDOWS.H. */
#ifndef SQL_TYPE_NULL
# define SQL_TYPE_NULL	0
#endif

/* --- module doc-string -------------------------------------------------- */

static char *Module_docstring = 

 MXODBC_MODULE" -- An ODBC interface. Version "VERSION"\n\n"

 "(c) Copyright Marc-Andre Lemburg, 1997, 1998, mal@lemburg.com,\n\n"
 "                 All Rights Reserved\n\n"
 "See the documentation for further information on copyrights,\n"
 "or contact the author."
;

/* --- module globals ----------------------------------------------------- */

static HENV henv;			/* ODBC Environment */
static PyObject *mxODBC_Error;		/* Base object for all other
					   exceptions */
static PyObject *mxODBC_InterfaceError;	/* Module Exception object */
static PyObject *mxODBC_InternalError;	/* ODBC Exception object */
static PyObject *mxODBC_OpError;	/* ODBC Exception object */
static PyObject *mxODBC_Warning;	/* ODBC Exception object */
static PyObject *mxODBC_ProgError;	/* ODBC Exception object */
static PyObject *mxODBC_IntegrityError;	/* ODBC Exception object */
static PyObject *mxODBC_DataError;	/* ODBC Exception object */
static PyObject *mxODBC_DatabaseError;	/* ODBC Exception object */
static PyObject *mxODBC_NotSuppError;	/* ODBC Exception object */
static PyObject *mxODBC_ErrorDict;	/* Dictionary mapping SQLState to
					   an exception object */

/* Intializer for the sqltypes dictionary and the type entries in the
   modules dict */

static mxODBC_TypeCodeTranslation mxODBC_TypeCodeTranslations[] = {
    {"CHAR",		SQL_CHAR},
    {"VARCHAR",		SQL_VARCHAR},
    {"LONGVARCHAR",	SQL_LONGVARCHAR},

    {"BINARY",		SQL_BINARY},
    {"VARBINARY",	SQL_VARBINARY},
    {"LONGVARBINARY",	SQL_LONGVARBINARY},

    {"TINYINT",		SQL_TINYINT},
    {"SMALLINT",        SQL_SMALLINT},
    {"INTEGER",		SQL_INTEGER},
    {"BIGINT",		SQL_BIGINT},
    {"DECIMAL",		SQL_DECIMAL},
    {"NUMERIC",		SQL_NUMERIC},
    {"BIT",		SQL_BIT},

    {"REAL",		SQL_REAL},
    {"FLOAT",		SQL_FLOAT},
    {"DOUBLE",		SQL_DOUBLE},

    {"DATE",		SQL_DATE},
    {"TIME",		SQL_TIME},
    {"TIMESTAMP",	SQL_TIMESTAMP},
};

/* Initializer for mxODBC_ErrorDict; ...no I did not enter this by
   hand :) The categories as well as the comments were adapted from
   the Win32 odbc.cpp file that comes with the source distribution of
   Mark Hammond's PythonWin for Python 1.4. Some of the assignements
   were subsequently changed to better fit the exception layout. */

static mxODBC_ErrorCodeTranslation mxODBC_ErrorCodeTranslations[] = {
    { "",      &mxODBC_OpError            }, /* Generic operational error */
    { "00000", &mxODBC_ProgError          }, /* Not documented */
    { "01000", &mxODBC_Warning            }, /* General warning */
    { "01002", &mxODBC_OpError            }, /* Disconnect error */
    { "01004", &mxODBC_Warning            }, /* Data truncated */
    { "01006", &mxODBC_InternalError      }, /* Privilege not revoked */
    { "01S00", &mxODBC_ProgError          }, /* Invalid connection string attribute */
    { "01S01", &mxODBC_InternalError      }, /* Error in row */
    { "01S02", &mxODBC_InternalError      }, /* Option value changed */
    { "01S03", &mxODBC_Warning            }, /* No rows updated or deleted */
    { "01S04", &mxODBC_Warning            }, /* More than one row updated or deleted */
    { "01S05", &mxODBC_Warning            }, /* Cancel treated as SQLFreeStmt with the SQL_CLOSE */
    { "01S06", &mxODBC_ProgError          }, /* Attempt to fetch before the result set returned */
    { "07001", &mxODBC_ProgError          }, /* Wrong number of parameters */
    { "07006", &mxODBC_ProgError          }, /* Restricted data type attribute violation */
    { "07S01", &mxODBC_ProgError          }, /* Invalid use of default parameter */
    { "08001", &mxODBC_OpError            }, /* Unable to connect to data source */
    { "08002", &mxODBC_OpError            }, /* Connection in use */
    { "08003", &mxODBC_OpError            }, /* Connection not open */
    { "08004", &mxODBC_OpError            }, /* Data source rejected establishment of connection */
    { "08007", &mxODBC_OpError            }, /* Connection failure during transaction */
    { "08S01", &mxODBC_OpError            }, /* Communication link failure */
    { "21S01", &mxODBC_ProgError          }, /* Insert value list does not match column list */
    { "21S02", &mxODBC_ProgError          }, /* Degree of derived table does not match column list */
    { "22001", &mxODBC_Warning            }, /* String data right truncation */
    { "22002", &mxODBC_InternalError      }, /* Indicator variable required but not supplied */
    { "22003", &mxODBC_DataError          }, /* Numeric value out of range */
    { "22005", &mxODBC_DataError          }, /* Error in assignment */
    { "22008", &mxODBC_DataError          }, /* Datetime field overflow */
    { "22012", &mxODBC_DataError          }, /* Division by zero */
    { "22026", &mxODBC_InternalError      }, /* String data, length mismatch */
    { "23000", &mxODBC_IntegrityError     }, /* Integrity constraint violation */
    { "24000", &mxODBC_InternalError      }, /* Invalid cursor state */
    { "25000", &mxODBC_InternalError      }, /* Invalid transaction state */
    { "28000", &mxODBC_OpError            }, /* Invalid authorization specification */
    { "34000", &mxODBC_InternalError      }, /* Invalid cursor name */
    { "37000", &mxODBC_ProgError          }, /* Syntax error or access violation */
    { "3C000", &mxODBC_InternalError      }, /* Duplicate cursor name */
    { "40001", &mxODBC_InternalError      }, /* Serialization failure */
    { "42000", &mxODBC_ProgError          }, /* Syntax error or access violation */
    { "70100", &mxODBC_OpError            }, /* Operation aborted */
    { "IM001", &mxODBC_NotSuppError       }, /* Driver does not support this function */
    { "IM002", &mxODBC_OpError            }, /* Data source name not found and no default driver  */
    { "IM003", &mxODBC_OpError            }, /* Specified driver could not be loaded */
    { "IM004", &mxODBC_OpError            }, /* Driver's SQLAllocEnv failed */
    { "IM005", &mxODBC_OpError            }, /* Driver's SQLAllocConnect failed */
    { "IM006", &mxODBC_OpError            }, /* Driver's SQLSetConnect-Option failed */
    { "IM007", &mxODBC_OpError            }, /* No data source or driver specified; dialog prohibited */
    { "IM008", &mxODBC_OpError            }, /* Dialog failed */
    { "IM009", &mxODBC_OpError            }, /* Unable to load translation DLL */
    { "IM010", &mxODBC_OpError            }, /* Data source name too long */
    { "IM011", &mxODBC_OpError            }, /* Driver name too long */
    { "IM012", &mxODBC_OpError            }, /* DRIVER keyword syntax error */
    { "IM013", &mxODBC_OpError            }, /* Trace file error */
    { "S0001", &mxODBC_ProgError          }, /* Base table or view already exists */
    { "S0002", &mxODBC_ProgError          }, /* Base table not found */
    { "S0011", &mxODBC_ProgError          }, /* Index already exists */
    { "S0012", &mxODBC_ProgError          }, /* Index not found */
    { "S0021", &mxODBC_ProgError          }, /* Column already exists */
    { "S0022", &mxODBC_ProgError          }, /* Column not found */
    { "S0023", &mxODBC_ProgError          }, /* No default for column */
    { "S1000", &mxODBC_OpError            }, /* General error */
    { "S1001", &mxODBC_OpError            }, /* Memory allocation failure */
    { "S1002", &mxODBC_InternalError      }, /* Invalid column number */
    { "S1003", &mxODBC_InternalError      }, /* Program type out of range */
    { "S1004", &mxODBC_InternalError      }, /* SQL data type out of range */
    { "S1008", &mxODBC_OpError            }, /* Operation canceled */
    { "S1009", &mxODBC_InternalError      }, /* Invalid argument value */
    { "S1010", &mxODBC_InternalError      }, /* Function sequence error */
    { "S1011", &mxODBC_InternalError      }, /* Operation invalid at this time */
    { "S1012", &mxODBC_InternalError      }, /* Invalid transaction operation code specified */
    { "S1015", &mxODBC_InternalError      }, /* No cursor name available */
    { "S1090", &mxODBC_InternalError      }, /* Invalid string or buffer length */
    { "S1091", &mxODBC_InternalError      }, /* Descriptor type out of range */
    { "S1092", &mxODBC_InternalError      }, /* Option type out of range */
    { "S1093", &mxODBC_InternalError      }, /* Invalid parameter number */
    { "S1095", &mxODBC_InternalError      }, /* Function type out of range */
    { "S1096", &mxODBC_InternalError      }, /* Information type out of range */
    { "S1097", &mxODBC_InternalError      }, /* Column type out of range */
    { "S1098", &mxODBC_InternalError      }, /* Scope type out of range */
    { "S1099", &mxODBC_InternalError      }, /* Nullable type out of range */
    { "S1100", &mxODBC_InternalError      }, /* Uniqueness option type out of range */
    { "S1101", &mxODBC_InternalError      }, /* Accuracy option type out of range */
    { "S1103", &mxODBC_InternalError      }, /* Direction option out of range */
    { "S1105", &mxODBC_InternalError      }, /* Invalid parameter type */
    { "S1106", &mxODBC_InternalError      }, /* Fetch type out of range */
    { "S1107", &mxODBC_InternalError      }, /* Row value out of range */
    { "S1108", &mxODBC_InternalError      }, /* Concurrency option out of range */
    { "S1109", &mxODBC_InternalError      }, /* Invalid cursor position */
    { "S1110", &mxODBC_InternalError      }, /* Invalid driver completion */
    { "S1111", &mxODBC_InternalError      }, /* Invalid bookmark value */
    { "S1C00", &mxODBC_NotSuppError       }, /* Driver not capable */
    { "S1T00", &mxODBC_OpError            }  /* Timeout expired */
};

/* --- module helpers ----------------------------------------------------- */

/* Create an exception object, insert it into the module dictionary
   under the given name and return the object pointer; this is NULL in
   case an error occurred. base can be given to indicate the base
   object to be used by the exception object. It should be NULL
   otherwise */

static 
PyObject *insexc(PyObject *moddict,
		 char *name,
		 PyObject *base)
{
    PyObject *v;
    char fullname[256];
    
    sprintf(fullname,MXODBC_MODULE".%s",name);
    v = PyErr_NewException(fullname, base, NULL);
    if (v == NULL)
	return NULL;
    if (PyDict_SetItemString(moddict,name,v))
	return NULL;
    return v;
}

/* In case an error occurred raise an appropriate exception and return
   -1; in case we had a false alarm (sqlstate == "00000") return 0 and
   don't raise an exception. 

   Also, when an exception is already set, don't overwrite it with a
   new one. Instead, log the error and return -1. */

#ifndef SQL_MAX_MESSAGE_LENGTH
# define SQL_MAX_MESSAGE_LENGTH 512
#endif

static
int mxODBC_ReportError(HENV henv, 
		       HDBC hdbc, 
		       HSTMT hstmt, 
		       int lineno, 
		       int orig_rc)
{
   UCHAR sqlstate[10];
   SDWORD sqlcode = 0;
   UCHAR text[SQL_MAX_MESSAGE_LENGTH];
   SWORD text_len;
   PyObject *v,*errorobj;
   RETCODE rc;

   rc = SQLError(henv, hdbc, hstmt, sqlstate, &sqlcode, text,
		 sizeof(text) - 1, &text_len);

   /* Check for false alarm: see ODBC docu Appendix A: Error codes */
   if (rc == SQL_NO_DATA_FOUND ||
       strcmp((char*)sqlstate,"00000") == 0)
       return 0;

#ifdef FLUSH_SQL_ERRORS
   do {
       UCHAR flush_sqlstate[10];
       UCHAR flush_text[SQL_MAX_MESSAGE_LENGTH];
       SWORD flush_text_len;
       SDWORD flush_sqlcode = 0;
       RETCODE flush_rc;

       flush_rc = SQLError(henv, hdbc, hstmt, 
			   flush_sqlstate, &flush_sqlcode, flush_text,
			   sizeof(flush_text) - 1, &flush_text_len);
       if (flush_rc == SQL_SUCCESS || flush_rc == SQL_SUCCESS_WITH_INFO) {
	   DPRINTF("mxODBC_ReportError: additional error info ignored "
		   "(sqlstate='%s' code=%i text='%s' lineno=%i rc=%i)\n",
		   flush_sqlstate,flush_sqlcode,flush_text,lineno,orig_rc);
       }
       else
	   break;
   } while (1);
#endif   

   /* Don't overwrite previous errors */
   if (PyErr_Occurred()) {
       if (rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO) {
	   DPRINTF("mxODBC_ReportError: error ignored "
		   "(sqlstate='%s' code=%i text='%s' lineno=%i rc=%i)\n",
		   sqlstate,sqlcode,text,lineno,orig_rc);
       }
       goto onError;
   }

   /* Check if SQLError() itself produced an error; truncations
      (SQL_SUCCESS_WITH_INFO) are not reported as errors. */
   if (rc == SQL_ERROR) {
       DPRINTF("mxODBC_ReportError: "
	       "failed to retrieve error information (line %i, rc=%i)\n",
	       lineno,orig_rc);
       PyErr_Format(mxODBC_InterfaceError,
		    "failed to retrieve error information (line %i, rc=%i)",
		    lineno,orig_rc);
       goto onError;
   }

   DPRINTF("mxODBC_ReportError: "
	   "sqlstate='%s' code=%i text='%s' lineno=%i rc=%i\n",
	   sqlstate,sqlcode,text,lineno,orig_rc);

   /* Go and find the appropriate error object */
   errorobj = PyDict_GetItemString(mxODBC_ErrorDict,(char *)sqlstate);
   if (errorobj == NULL)
       errorobj = mxODBC_InterfaceError;
   v = Py_BuildValue("(sis#i)",sqlstate,sqlcode,text,text_len,lineno);
   if (v != NULL) {
       PyErr_SetObject(errorobj, v);
       Py_DECREF(v);
   }

 onError:
   return -1;
}

/* Macros for handling SQL errors.

   The return code is expected in variable rc; handles not available
   can be set to NULL;

   The compile time option DONT_REPORT_WARNINGS can be defined
   to disable reporting SQL Warnings (those that return with
   SQL_SUCCESS_WITH_INFO as return code).

   XXX Don't like these many different macros...

*/

/* The mxODBC_SQLCheck macro reports all errors *and* warnings per
   default. Defining DONT_REPORT_WARNINGS turns reporting of warnings
   off.  Always use this interface unless you are sure that a function
   will always return SQL_SUCCESS_WITH_INFO -- revert to
   mxODBC_SQLErrorCheck() in that case. */

#ifdef DONT_REPORT_WARNINGS

# define mxODBC_SQLCheck(henv,hdbc,hstmt)				\
    if (rc != SQL_SUCCESS) {						\
	if (mxODBC_ReportError(henv,hdbc,hstmt,__LINE__,(int)rc))	\
	    goto onError;						\
    }

#else

# define mxODBC_SQLCheck(henv,hdbc,hstmt)				\
    if (rc != SQL_SUCCESS && rc != SQL_SUCCESS_WITH_INFO) {		\
	if (mxODBC_ReportError(henv,hdbc,hstmt,__LINE__,(int)rc))	\
	    goto onError;						\
    }

#endif

/* This macro only reports errors. Warnings are always ignored. */

#define mxODBC_SQLErrorCheck(henv,hdbc,hstmt)				\
    if (rc != SQL_SUCCESS && rc != SQL_SUCCESS_WITH_INFO) {		\
	if (mxODBC_ReportError(henv,hdbc,hstmt,__LINE__,(int)rc))	\
	    goto onError;						\
    }


/* This macro records errors only. It does not jump to onError but
   does set an Python exception. Check with PyErr_Occurred()
   afterwards. */

#define mxODBC_SQLErrorCheck_Only(henv,hdbc,hstmt)		\
    if (rc != SQL_SUCCESS && rc != SQL_SUCCESS_WITH_INFO) {	\
	mxODBC_ReportError(henv,hdbc,hstmt,__LINE__,(int)rc);	\
    }

/* Helper for doing rollbacks. Performs rollbacks only after having
   checked for the driver capability. An exception is raised and -1
   returned in case the rollback fails or is not supported by the data
   source; 0 is returned otherwise. */

#ifndef DONT_HAVE_TRANSACTIONS
static
int mxODBC_Rollback(mxODBCObject *dbc)
{
    RETCODE rc;
    char value[16];
    SWORD value_len;

    Py_Assert(!dbc->closed,
	      mxODBC_ProgError,
	      "connection already closed");
    
    DPRINTF("mxODBC_Rollback: Rolling back changes for connection at 0x%lx\n",
	    (long)dbc);

    /* Check whether the connection supports transactions. */
    rc = SQLGetInfo(dbc->hdbc,
		    SQL_TXN_CAPABLE,
		    (PTR)value,
		    (SWORD)sizeof(value),
		    &value_len);
    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
    Py_Assert(*(UWORD*)value != SQL_TC_NONE,
	      mxODBC_NotSuppError,
	      "connection does not support transactions");

    /* Check AUTOCOMMIT */
    rc = SQLGetConnectOption(dbc->hdbc,
			     SQL_AUTOCOMMIT,
			     (PTR)value);
    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
    Py_Assert(*(UWORD*)value == SQL_AUTOCOMMIT_OFF,
	      mxODBC_NotSuppError,
	      "connection has AUTOCOMMIT turned on");
    
    /* Execute the roll back */
    Py_BEGIN_ALLOW_THREADS;
    rc = SQLTransact(henv,dbc->hdbc,SQL_ROLLBACK);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);

    return 0;
    
 onError:
    return -1;
}
#endif

/* --- Database Cursor Object ------------------------------------------*/

staticforward PyMethodDef mxODBCursor_Methods[];

/* --- allocation --- */

/* Create a new cursor object;

   If name is NULL, no explicit name will be set for the cursor.
*/

static
PyObject *mxODBCursor_New(mxODBCObject *dbc,
			  char *name)
{
    mxODBCursorObject *dbcs;
    RETCODE rc;
    int allocated = 0;

    if (dbc->closed) {
	PyErr_SetString(mxODBC_ProgError,
			"connection already closed");
	return NULL;
    }

    dbcs = PyObject_NEW(mxODBCursorObject,&mxODBCursor_Type);
    if (dbcs == NULL)
	return NULL;

    /* Get a hold of the connection object */
    dbcs->dbc = dbc;
    Py_INCREF(dbcs->dbc);

    /* Alloc Statement  */
    rc = SQLAllocStmt(dbc->hdbc, &dbcs->hstmt);
    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
    allocated = 1;

    /* Init vars */
    dbcs->colcount = -1;
    dbcs->rowcount = -1;
    dbcs->vars = 0;
    dbcs->sqlcmd = 0;
    dbcs->arraysize = 1;
    dbcs->datetimeformat = dbc->datetimeformat;
    dbcs->closed = 0;

    /* Set cursor name */
    if (name) {
	rc = SQLSetCursorName(dbcs->hstmt,
			      (UCHAR *)name,
			      (SWORD)strlen(name));
	mxODBC_SQLCheck(henv,dbc->hdbc,dbcs->hstmt);
    }

    DPRINTF("mxODBCursor_New: created new cursor '%s' at 0x%lx\n",
	    name == NULL ? "<no name>" : name, (long)dbcs);
    return (PyObject *)dbcs;

 onError:
    if (allocated)
	SQLFreeStmt(dbcs->hstmt,SQL_DROP);
    /* We don't want the deallocator to be called... */
    PyMem_DEL(dbcs);
    /* This may call the connection deallocator and overwrite our error
       message... oh well. */
    Py_DECREF(dbc);
    return NULL;
}

/* --- deallocation --- */

static
int mxODBCursor_Close(mxODBCursorObject *dbcs)
{
    RETCODE rc = SQL_SUCCESS;
    
    DPRINTF("mxODBCursor_Close: called for cursor at 0x%lx\n",(long)dbcs);
    if (!dbcs->closed) {

	/* "Forget" the last executed SQL statement */
	if (dbcs->sqlcmd) {
	    Py_DECREF(dbcs->sqlcmd);
	    dbcs->sqlcmd = NULL;
	}
	dbcs->closed = 1; 

	/* Cancel any ongoing action and drop the cursor; errors are
	   *not* reported to the user. Doing this is only safe if the
	   connection still is alive. It may otherwise cause a
	   core dump with some drivers. */
	if (!dbcs->dbc->closed) {
	    rc = SQLCancel(dbcs->hstmt);
	    IF_DEBUGGING {
		if (rc != SQL_SUCCESS) {
		    DPRINTF("mxODBCursor_Close:   error cancelling; rc=%i",
			    rc);
		}
		else {
		    DPRINTF("mxODBCursor_Close:   stmt cancelled\n");
		}
	    }
	    rc = SQLFreeStmt(dbcs->hstmt,SQL_DROP);
	    if (rc != SQL_SUCCESS) {
		DPRINTF("mxODBCursor_Close:   error freeing stmt; rc=%i",
			rc);
	    }
	    else {
		DPRINTF("mxODBCursor_Close:   stmt freed\n");
	    }
	}
	else {
	    DPRINTF("mxODBCursor_Close:   connection already closed; "
		    "stmt *not* cancelled or freed !\n");
	}
    }
    return 0;
}

/* Free any allocated parameter variables and unbind them. */

static
void mxODBCursor_FreeParameters(mxODBCursorObject *dbcs)
{
    RETCODE rc;

    DPRINTF("mxODBCursor_FreeParameters: called for cursor at 0x%lx\n",
	    (long)dbcs);

    /* Nothing to do: the cursor is already closed or the connection
       gone. */
    if (dbcs->closed || dbcs->dbc->closed) {
	DPRINTF("mxODBCursor_FreeParameters:  cursor or connection closed !\n");
	return;
    }

    /* Unbind all possibly still bound parameters */
    rc = SQLFreeStmt(dbcs->hstmt,SQL_RESET_PARAMS);
    IF_DEBUGGING {
	if (rc != SQL_SUCCESS)
	    DPRINTF("mxODBCursor_FreeVars:  "
		    "parameter unbind returned an error: rc=%i",
		    rc);
    }

}

/* Free any allocated output variables in the dbcs->vars array plus
   the array itself
   - Important note: since dbcs->colcount indicates the size of the
     array, this variable must *not* be changed after the array is
     allocated !!!
*/

static
void mxODBCursor_FreeVars(mxODBCursorObject *dbcs)
{
    RETCODE rc;

    DPRINTF("mxODBCursor_FreeVars: called for cursor at 0x%lx\n",(long)dbcs);

    if (!dbcs->closed && !dbcs->dbc->closed) {
	/* Unbind all possibly still bound variables */
	rc = SQLFreeStmt(dbcs->hstmt,SQL_UNBIND);
	IF_DEBUGGING {
	    if (rc != SQL_SUCCESS)
		DPRINTF("mxODBCursor_FreeVars:  "
			"variable unbind returned an error: rc=%i",
			rc);
	}
    }
    else {
	DPRINTF("mxODBCursor_FreeVars:  cursor or connection closed !\n");
    }
    
    /* Free output vars array */
    if (dbcs->vars) {
	mxODBCursor_Variable *vars = dbcs->vars;
	int i,n = dbcs->colcount;

	/* Release allocated variable storage */
	for (i = 0; i < n; i++) {
	    if (vars[i].free_data && vars[i].data) {
		DPRINTF("mxODBCursor_FreeVars:  "
			"freeing output variable for column %i\n",i);
		free(vars[i].data);
	    }
	}
	/* Release variable array */
	DPRINTF("mxODBCursor_FreeVars:  "
		"freeing output variable array\n");
	free(vars);
	dbcs->vars = 0;
    }
}

static
void mxODBCursor_Free(mxODBCursorObject *dbcs)
{
    DPRINTF("mxODBCursor_Free: called for cursor at 0x%lx\n",(long)dbcs);
    mxODBCursor_FreeVars(dbcs);
    mxODBCursor_FreeParameters(dbcs);
    if (mxODBCursor_Close(dbcs))
	/* There's no way to report this error */
	PyErr_Clear();
    Py_DECREF(dbcs->dbc);
    PyMem_DEL(dbcs);
}

/* Reset the cursor, e.g. to issue SQLTables() etc. afterwards */

static
int mxODBCursor_Reset(mxODBCursorObject *dbcs)
{
    RETCODE rc;
    DPRINTF("mxODBCursor_Reset: called for cursor at 0x%lx\n",(long)dbcs);

    /* Cancel any ongoing action and close the cursor. Doing this
       is only safe if the connection still is alive. It might
       otherwise cause a core dump. */
    if (!dbcs->dbc->closed) {
	/* Free cursor variables and parameters */
	mxODBCursor_FreeVars(dbcs);
	mxODBCursor_FreeParameters(dbcs);
	/* Cancel and free the last cursor statement */
	rc = SQLCancel(dbcs->hstmt);
	mxODBC_SQLErrorCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
	rc = SQLFreeStmt(dbcs->hstmt,SQL_CLOSE);
	mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
	DPRINTF("mxODBCursor_Reset:   stmt cancelled and cursor closed\n");
    }
    
    /* "Forget" the last executed SQL statement */
    if (dbcs->sqlcmd) {
	Py_DECREF(dbcs->sqlcmd);
	dbcs->sqlcmd = NULL;
    }
    return 0;
    
 onError:
    return -1;
}

/* --- internal functions --- */

/* After an execute, this function returns a tuple indicating
   the column formats of each result column in a 7-tuple. It
   return NULL on error.
*/

static
PyObject *mxODBCursor_Description(mxODBCursorObject *dbcs)
{
    PyObject *t = 0;
    int i;
    mxODBCursor_Variable *vars = dbcs->vars;

    if (dbcs->colcount <= 0)
	Py_ReturnNone();
    
    t = PyTuple_New(dbcs->colcount);
    if (!t)
	goto onError;

    for (i = 0; i < dbcs->colcount; i++) {
	PyObject *v;

	v = Py_BuildValue("(s#iOOiii)",
			  vars[i].colname,
			  vars[i].colname_len,
			  vars[i].sqltype,
			  Py_None,Py_None,
			  vars[i].sqllen,
			  vars[i].sqlscale,
			  vars[i].nullable);
	if (!v)
	    goto onError;
	PyTuple_SET_ITEM(t,i,v);
    }

    return t;

 onError:
    Py_XDECREF(t);
    return NULL;
}

/* --------------------------------------------------------------------------
   Converters from SQL C types to Python types 

   These may raise exceptions and then return NULL to indicate this.

*/

/* Python string from C string (with size). The string may contain
   embedded \0 characters. */

static
PyObject *mxODBC_StringFromString(mxODBCursor_Variable *var)
{
    return PyString_FromStringAndSize((char *)var->data,var->data_len);
}

/* Python long from 0-terminated C string. */

static
PyObject *mxODBC_LongFromString(mxODBCursor_Variable *var)
{
    /* Note: The string has to be 0-terminated. */
    return PyLong_FromString((char *)var->data, (char **)NULL, 10);
}

/* Python integer from C long. */

static
PyObject *mxODBC_IntFromLong(mxODBCursor_Variable *var)
{
    return PyInt_FromLong(*(long *)var->data);
}

/* Python float from C double. */

static
PyObject *mxODBC_FloatFromDouble(mxODBCursor_Variable *var)
{
    return PyFloat_FromDouble(*(double *)var->data);
}

/* TIMESTAMP is converted to a DateTime instance */

static
PyObject *mxODBC_DateTimeFromTimestampStruct(mxODBCursor_Variable *var)
{
    TIMESTAMP_STRUCT *time = (TIMESTAMP_STRUCT *)var->data;

    DPRINTF("mxODBC_DateTimeFromTimestampStruct: "
	    "year=%i month=%i day=%i hour=%i minute=%i second=%i fraction=%i\n",
	    time->year,time->month,time->day,time->hour,
	    time->minute,time->second,time->fraction);

    return mxDateTime.DateTime_FromDateAndTime(time->year,
					       time->month,
					       time->day,
					       time->hour,
					       time->minute,
					       (double)time->second +
					       (double)time->fraction / 
					               FRACTIONS_IN_SECOND);
}

/* TIMESTAMP is converted to a seconds since epoch (01.01.70, 00:00:00.00)
   double */

static
PyObject *mxODBC_FloatFromTimestampStruct(mxODBCursor_Variable *var)
{
    TIMESTAMP_STRUCT *time = (TIMESTAMP_STRUCT *)var->data;
    double ticks;
    struct tm tm;

    memset(&tm,0,sizeof(tm));

    tm.tm_sec = time->second;
    tm.tm_min = time->minute;
    tm.tm_hour = time->hour;
    tm.tm_mday = time->day;
    tm.tm_mon = time->month - 1;
    tm.tm_year = time->year - 1900;
    tm.tm_isdst = -1;
    ticks = (double)mktime(&tm);
    Py_Assert(ticks != -1.0,
	      PyExc_OverflowError,
	      "can't convert to TIMESTAMP to ticks float");
    ticks += (double)time->fraction / FRACTIONS_IN_SECOND;
    return PyFloat_FromDouble(ticks);
 onError:
    return NULL;
}

/* TIMESTAMP is converted to a tuple (year,month,day,hour,minute,second) */

static
PyObject *mxODBC_TupleFromTimestampStruct(mxODBCursor_Variable *var)
{
    TIMESTAMP_STRUCT *time = (TIMESTAMP_STRUCT *)var->data;
    PyObject *v;

    v = Py_BuildValue("iiiiid",
		      time->year,
		      time->month,
		      time->day,
		      time->hour,
		      time->minute,
		      (double)time->second + 
		      (double)time->fraction / FRACTIONS_IN_SECOND);
    return v;
}

/* DATE is converted to a DateTime instance using 00:00:00.00 as time */

static
PyObject *mxODBC_DateTimeFromDateStruct(mxODBCursor_Variable *var)
{
    DATE_STRUCT *time = (DATE_STRUCT *)var->data;

    return mxDateTime.DateTime_FromDateAndTime(time->year,
					       time->month,
					       time->day,
					       0,
					       0,
					       0.0);
}

/* DATE is converted to a seconds since epoch double using 00:00:00.00
   as time
*/

static
PyObject *mxODBC_FloatFromDateStruct(mxODBCursor_Variable *var)
{
    DATE_STRUCT *time = (DATE_STRUCT *)var->data;
    double ticks;
    struct tm tm;

    memset(&tm,0,sizeof(tm));

    tm.tm_mday = time->day;
    tm.tm_mon = time->month - 1;
    tm.tm_year = time->year - 1900;
    tm.tm_isdst = -1;
    ticks = (double)mktime(&tm);
    Py_Assert(ticks != -1.0,
	      PyExc_OverflowError,
	      "can't convert to TIMESTAMP to ticks float");
    return PyFloat_FromDouble(ticks);
 onError:
    return NULL;
}

/* DATE is converted to a tuple (year,month,day) */

static
PyObject *mxODBC_TupleFromDateStruct(mxODBCursor_Variable *var)
{
    DATE_STRUCT *time = (DATE_STRUCT *)var->data;
    PyObject *v;

    v = Py_BuildValue("iii",
		      time->year,
		      time->month,
		      time->day);
    return v;
}

/* TIME is converted to a DateTimeDelta instance */

static
PyObject *mxODBC_DateTimeDeltaFromTimeStruct(mxODBCursor_Variable *var)
{
    TIME_STRUCT *time = (TIME_STRUCT *)var->data;
    
    return mxDateTime.DateTimeDelta_FromTime(time->hour,
					     time->minute,
					     (double)time->second);
}

/* TIME is converted to a seconds since 00:00:00.00 double */

static
PyObject *mxODBC_FloatFromTimeStruct(mxODBCursor_Variable *var)
{
    TIME_STRUCT *time = (TIME_STRUCT *)var->data;
    double ticks;
    ticks = time->hour * 3600.0 + time->minute * 60.0 + (double)time->second;
    return PyFloat_FromDouble(ticks);
}

/* TIME is converted to a tuple (hour,minute,seconds) */

static
PyObject *mxODBC_TupleFromTimeStruct(mxODBCursor_Variable *var)
{
    TIME_STRUCT *time = (TIME_STRUCT *)var->data;
    PyObject *v;

    v = Py_BuildValue("iii",
		      time->hour,
		      time->minute,
		      time->second);
    return v;
}

/* --------------------------------------------------------------------------
   Converters from Python types to SQL C types.

   The functions may set exceptions and return -1 or just return -1;
   the callee has to deal with this properly.

   The functions may allocate memory for var->data which has to be
   freed again after use (var->free_data is set to 1 to indicate
   this).  The allocated storage size has to be stored in
   var->max_len, so that subsequent calls can reuse the already
   allocated storage.
   
*/


/* Convert an object to a string */

static
int mxODBC_StringFromObject(PyObject *v,
			    mxODBCursor_Variable *var)
{
    char *str;
    PyObject *s = 0;
    int size;

    DPRINTF("mxODBC_StringFromObject: called for object %s at 0x%x\n",
	    v->ob_type->tp_name,(int)v);

    /* Special case for DateTimeDelta instances: */
    if (mxDateTimeDelta_Check(v)) {
	long day;
	int hour;
	int minute;
	double second;

	mxDateTime.DateTimeDelta_BrokenDown((mxDateTimeDeltaObject *)v,
					    &day,&hour,&minute,&second);
	Py_Assert(day == 0,
		  mxODBC_InterfaceError,
		  "DateTimeDelta instances with day part not supported");

	size = 8;

	str = new(char,size+1);
	if (!str) {
	    PyErr_NoMemory();
	    goto onError;
	}
	var->free_data = 1;
	var->max_len = size + 1;
	sprintf(str,"%02i:%02i:%02i",hour,minute,(int)second);
    }

#ifdef HAVE_PYTHON_BUFFEROBJECTS
    /* Try buffer conversion; available in Python >= 1.5.2 */
    else if (PyBuffer_Check(v)) {
	const char *const_str;
	
	if (PyObject_AsCharBuffer(v,&const_str,&size))
	    goto onError;
	str = (char *)const_str;
    }
#endif

    /* Generic string conversion: */
    else {
	s = PyObject_Str(v);
	if (!s)
	    goto onError;
	Py_Assert(PyString_Check(s),
		  PyExc_TypeError,
		  "str(object) must return a Python string");

	size = PyString_GET_SIZE(s);

	/* Also copy the terminating \0 byte... */
	str = new(char,size + 1);
	if (!str) {
	    PyErr_NoMemory();
	    goto onError;
	}
	var->free_data = 1;
	var->max_len = size + 1;
	memcpy(str, PyString_AS_STRING(s), size + 1);
	Py_DECREF(s);
    }

    var->data = (PTR)str;
    var->data_len = size;
    var->ctype = SQL_C_CHAR;
    return 0;
    
 onError:
    size = 0;
    Py_XDECREF(s);
    return -1;
}

/* Convert an object to a binary field */

static
int mxODBC_BinaryFromObject(PyObject *v,
			    mxODBCursor_Variable *var)
{
    DPRINTF("mxODBC_BinaryFromObject: called for object %s at 0x%x\n",
	    v->ob_type->tp_name,(int)v);

    /* Do the same conversion as for string data and then set the type
       to SQL_C_BINARY. */
    if (mxODBC_StringFromObject(v,var))
	goto onError;
    DPRINTF("mxODBC_BinaryFromObject: setting C type to SQL_C_BINARY\n");
    var->ctype = SQL_C_BINARY;
    return 0;
    
 onError:
    return -1;
}

/* Convert a Python long to a string */

static
int mxODBC_StringFromLongObject(PyObject *v,
				mxODBCursor_Variable *var)
{
    char *str;
    PyObject *s = 0;
    int size;

    DPRINTF("mxODBC_StringFromLongObject: called for object %s at 0x%x\n",
	    v->ob_type->tp_name,(int)v);

    Py_Assert(PyLong_Check(v),
	      PyExc_TypeError,
	      "need a Python long");
	      
    s = PyObject_Str(v);
    if (!s)
	goto onError;
    Py_Assert(PyString_Check(s),
	      PyExc_SystemError,
	      "str(long) dir not return a Python string");

    size = PyString_GET_SIZE(s);
    Py_Assert(PyString_AS_STRING(s)[size - 1] == 'L',
	      mxODBC_InterfaceError,
	      "mxODBC_StringFromLongObject: str(long) format changed");

    str = new(char,size);
    if (!str) {
	PyErr_NoMemory();
	goto onError;
    }
    memcpy(str, PyString_AS_STRING(s), size);
    Py_DECREF(s);

    /* Chop off the trailing 'L' and make the string 0-terminated again */
    size--;
    str[size] = '\0';

    var->data = (PTR)str;
    var->free_data = 1;
    var->max_len = size;
    var->data_len = size;
    var->ctype = SQL_C_CHAR;
    return 0;
    
 onError:
    Py_XDECREF(s);
    return -1;
}

static
int mxODBC_LongFromObject(PyObject *v,
			  mxODBCursor_Variable *var)
{
    long *valueptr;
    long value;
    PyObject *w = 0;

    DPRINTF("mxODBC_LongFromObject: called for object %s at 0x%x\n",
	    v->ob_type->tp_name,(int)v);

    if (PyString_Check(v)) {
	char *end;

	/* We mimic the int(str) behaviour here: */
	Py_Assert((int)strlen(PyString_AS_STRING(v)) == PyString_GET_SIZE(v),
		  PyExc_TypeError,
		  "expected string without null bytes");
	errno = 0;
	value = PyOS_strtol(PyString_AS_STRING(v),&end,10);
	Py_Assert(errno == 0,
		  PyExc_ValueError,
		  "literal too large for integer conversion");
	while (*end && isspace(Py_CHARMASK(*end)))
	    end++;
	Py_Assert(*end == '\0',
		  PyExc_ValueError,
		  "invalid literal for integer conversion");
    }
    else {
	/* Try the __int__ method... */
	w = PyNumber_Int(v);
	if (!w)
	    goto onError;
	Py_Assert(PyInt_Check(w),
		  PyExc_TypeError,
		  "int(object) must return a Python integer");
	value = PyInt_AS_LONG(w);
	Py_DECREF(w);
    }

    valueptr = new(long,1);
    if (!valueptr) {
	PyErr_NoMemory();
	goto onError;
    }
    *valueptr = value;

    var->data = (PTR)valueptr;
    var->free_data = 1;
    var->max_len = sizeof(long);
    var->data_len = sizeof(long);
    var->ctype = SQL_C_SLONG;
    return 0;
    
 onError:
    Py_XDECREF(w);
    return -1;
}

static
int mxODBC_DoubleFromObject(PyObject *v,
			    mxODBCursor_Variable *var)
{
    double *valueptr;
    double value;
    PyObject *w = 0;

    DPRINTF("mxODBC_DoubleFromObject: called for object %s at 0x%x\n",
	    v->ob_type->tp_name,(int)v);

    if (PyString_Check(v)) {
	char *end;

	/* We mimic the float(str) behaviour here: */
	Py_Assert((int)strlen(PyString_AS_STRING(v)) == PyString_GET_SIZE(v),
		  PyExc_TypeError,
		  "expected string without null bytes");
	errno = 0;
	PyFPE_START_PROTECT("mxODBC_DoubleFromObject", return -1)
	value = strtod(PyString_AS_STRING(v),&end);
	PyFPE_END_PROTECT(value)
	Py_Assert(errno == 0,
		  PyExc_ValueError,
		  "literal too large for float conversion");
	while (*end && isspace(Py_CHARMASK(*end)))
	    end++;
	Py_Assert(*end == '\0',
		  PyExc_ValueError,
		  "invalid literal for float conversion");
    }
    else {
	/* Try the __float__ method... */
	w = PyNumber_Float(v);
	if (!w)
	    goto onError;
	Py_Assert(PyFloat_Check(w),
		  PyExc_TypeError,
		  "float(object) must return a Python float");
	value = PyFloat_AS_DOUBLE(w);
	Py_DECREF(w);
    }

    valueptr = new(double,1);
    if (!valueptr) {
	PyErr_NoMemory();
	goto onError;
    }
    *valueptr = value;

    var->data = (PTR)valueptr;
    var->free_data = 1;
    var->max_len = sizeof(double);
    var->data_len = sizeof(double);
    var->ctype = SQL_C_DOUBLE;
    return 0;
    
 onError:
    Py_XDECREF(w);
    return -1;
}

static
int mxODBC_DateStructFromDateTime(PyObject *v,
				  mxODBCursor_Variable *var)
{
    DATE_STRUCT *date = 0;
    long year;
    int month,day;

#if 0
    Py_Assert(mxDateTime_Check(v),
	      PyExc_TypeError,
	      "date must be a DateTime instance");
#endif

    date = newstruct(DATE_STRUCT);
    if (!date) {
	PyErr_NoMemory();
	goto onError;
    }
    
    if (mxDateTime.DateTime_BrokenDown((mxDateTimeObject *)v,
				       &year,
				       &month,
				       &day,
				       NULL,NULL,NULL))
	goto onError;
    date->year = (int)year;
    date->month = month;
    date->day = day;

    var->data = (PTR)date;
    var->data_len = sizeof(DATE_STRUCT);
    var->max_len = sizeof(DATE_STRUCT);
    var->free_data = 1;
    var->ctype = SQL_C_DATE;
    return 0;
    
 onError:
    if (date)
	free(date);
    return -1;
}

static
int mxODBC_DateStructFromFloat(PyObject *v,
			       mxODBCursor_Variable *var)
{
    DATE_STRUCT *date;
    time_t ticks;
    struct tm *tm;

    ticks = (time_t)PyFloat_AsDouble(v);
    if (!ticks)
	goto onError;

    tm = localtime(&ticks);
    date = newstruct(DATE_STRUCT);
    if (!date) {
	PyErr_NoMemory();
	goto onError;
    }
    date->year = 1900 + tm->tm_year;
    date->month = 1 + tm->tm_mon;
    date->day = tm->tm_mday;

    var->data = (PTR)date;
    var->max_len = sizeof(DATE_STRUCT);
    var->data_len = sizeof(DATE_STRUCT);
    var->free_data = 1;
    var->ctype = SQL_C_DATE;
    return 0;
    
 onError:
    return -1;
}

static
int mxODBC_DateStructFromTuple(PyObject *v,
			       mxODBCursor_Variable *var)
{
    DATE_STRUCT *date;
    int year,month,day;

#if 0
    if (!PyTuple_Check(v))
	goto onError;
#endif

    if (!PyArg_ParseTuple(v,"iii;need a date 3-tuple (year,month,day)",
			  &year,&month,&day))
	goto onError;

    /* Range checks */
    Py_Assert(year >= 0 && year <= 9999,
	      PyExc_ValueError,
	      "year must be in the range 0-9999");
    Py_Assert(month >= 1 && month <= 12,
	      PyExc_ValueError,
	      "month must be in the range 1-12");
    Py_Assert(day >= 1 && day <= 31,
	      PyExc_ValueError,
	      "day must be in the range 1-31");
    
    date = newstruct(DATE_STRUCT);
    if (!date) {
	PyErr_NoMemory();
	goto onError;
    }

    date->year = year;
    date->month = month;
    date->day = day;

    var->data = (PTR)date;
    var->max_len = sizeof(DATE_STRUCT);
    var->data_len = sizeof(DATE_STRUCT);
    var->free_data = 1;
    var->ctype = SQL_C_DATE;
    return 0;
    
 onError:
    return -1;
}

static
int mxODBC_TimeStructFromDateTimeDelta(PyObject *v,
				       mxODBCursor_Variable *var)
{
    TIME_STRUCT *time = 0;
    int hour,minute;
    double second;

    Py_Assert(mxDateTimeDelta_Check(v),
	      PyExc_TypeError,
	      "time must be a DateTimeDelta instance");

    time = newstruct(TIME_STRUCT);
    if (!time) {
	PyErr_NoMemory();
	goto onError;
    }

    if (mxDateTime.DateTimeDelta_BrokenDown((mxDateTimeDeltaObject *)v,
					    NULL,
					    &hour,
					    &minute,
					    &second))
	goto onError;
    time->hour  = hour;
    time->minute = minute;
    time->second   = (int)second;

    var->data = (PTR)time;
    var->max_len = sizeof(TIME_STRUCT);
    var->data_len = sizeof(TIME_STRUCT);
    var->free_data = 1;
    var->ctype = SQL_C_TIME;
    return 0;
    
 onError:
    if (time)
	free(time);
    return -1;
}

static
int mxODBC_TimeStructFromFloat(PyObject *v,
			       mxODBCursor_Variable *var)
{
    TIME_STRUCT *time;
    time_t ticks;

    ticks = (time_t)PyFloat_AsDouble(v);
    if (!ticks)
	goto onError;
    
    time = newstruct(TIME_STRUCT);
    if (!time) {
	PyErr_NoMemory();
	goto onError;
    }
    time->hour  = ticks / 3600; ticks %= 3600;
    time->minute = ticks / 60;
    time->second   = ticks % 60;

    var->data = (PTR)time;
    var->max_len = sizeof(TIME_STRUCT);
    var->data_len = sizeof(TIME_STRUCT);
    var->free_data = 1;
    var->ctype = SQL_C_TIME;
    return 0;
    
 onError:
    return -1;
}

static
int mxODBC_TimeStructFromTuple(PyObject *v,
			       mxODBCursor_Variable *var)
{
    TIME_STRUCT *time;
    int hour,minute,second;

    if (!PyArg_ParseTuple(v,"iii;need a time 3-tuple (hour,min,sec)",
			  &hour,&minute,&second))
	goto onError;

    /* Range checks */
    Py_Assert(hour >= 0 && hour <= 23,
	      PyExc_ValueError,
	      "hour must be in the range 0-23");
    Py_Assert(minute >= 0 && minute <= 59,
	      PyExc_ValueError,
	      "minute must be in the range 0-59");
    Py_Assert(second >= 0 && second <= 59,
	      PyExc_ValueError,
	      "second must be in the range 0-59");
    
    time = newstruct(TIME_STRUCT);
    if (!time) {
	PyErr_NoMemory();
	goto onError;
    }
    time->hour  = hour;
    time->minute = minute;
    time->second   = second;

    var->data = (PTR)time;
    var->max_len = sizeof(TIME_STRUCT);
    var->data_len = sizeof(TIME_STRUCT);
    var->free_data = 1;
    var->ctype = SQL_C_TIME;
    return 0;
    
 onError:
    return -1;
}

static
int mxODBC_TimestampStructFromDateTime(PyObject *v,
				       mxODBCursor_Variable *var)
{
    TIMESTAMP_STRUCT *time = 0;
    long year;
    int month,day,hour,minute;
    double second;

    Py_Assert(mxDateTime_Check(v),
	      PyExc_TypeError,
	      "timestamp must be a DateTime instance");

    time = newstruct(TIMESTAMP_STRUCT);
    if (!time) {
	PyErr_NoMemory();
	goto onError;
    }

    if (mxDateTime.DateTime_BrokenDown((mxDateTimeObject *)v,
				       &year,
				       &month,
				       &day,
				       &hour,
				       &minute,
				       &second))
	goto onError;
    time->year = (SWORD)year;
    time->month = month;
    time->day = day;
    time->hour  = hour;
    time->minute = minute;
    time->second   = (UWORD)second;
    time->fraction = (UDWORD)((second - (double)((int)second))
			      * FRACTIONS_IN_SECOND);

    DPRINTF("mxODBC_TimestampStructFromDateTime: "
	    "year=%i month=%i day=%i hour=%i minute=%i second=%i fraction=%i\n",
	    time->year,time->month,time->day,time->hour,
	    time->minute,time->second,time->fraction);

    var->data = (PTR)time;
    var->max_len = sizeof(TIMESTAMP_STRUCT);
    var->data_len = sizeof(TIMESTAMP_STRUCT);
    var->free_data = 1;
    var->ctype = SQL_C_TIMESTAMP;
    return 0;
    
 onError:
    if (time)
	free(time);
    return -1;
}

static
int mxODBC_TimestampStructFromFloat(PyObject *v,
				    mxODBCursor_Variable *var)
{
    TIMESTAMP_STRUCT *time;
    double fticks;
    time_t ticks;
    struct tm *tm;

    fticks = PyFloat_AsDouble(v);
    if (!fticks)
	goto onError;
    ticks = (time_t)fticks;

    tm = localtime(&ticks);
    time = newstruct(TIMESTAMP_STRUCT);
    if (!time) {
	PyErr_NoMemory();
	goto onError;
    }
    time->hour  = tm->tm_hour;
    time->minute = tm->tm_min;
    time->second   = tm->tm_sec;
    time->year = 1900 + tm->tm_year;
    time->month = 1 + tm->tm_mon;
    time->day = tm->tm_mday;
    time->fraction = (UDWORD)((fticks - (double)ticks) * FRACTIONS_IN_SECOND);

    var->data = (PTR)time;
    var->max_len = sizeof(TIMESTAMP_STRUCT);
    var->data_len = sizeof(TIMESTAMP_STRUCT);
    var->free_data = 1;
    var->ctype = SQL_C_TIMESTAMP;
    return 0;
    
 onError:
    return -1;
}

static
int mxODBC_TimestampStructFromTuple(PyObject *v,
				    mxODBCursor_Variable *var)
{
    TIMESTAMP_STRUCT *time;
    int year,month,day,hour,minute;
    double second;

    if (!PyArg_ParseTuple(v,"iiiiid;need a timestamp 6-tuple",
			  &year,&month,&day,&hour,
			  &minute,&second))
	goto onError;

    /* Range checks */
    Py_Assert(year >= 0 && year <= 9999,
	      PyExc_ValueError,
	      "year must be in the range 0-9999");
    Py_Assert(month >= 1 && month <= 12,
	      PyExc_ValueError,
	      "month must be in the range 1-12");
    Py_Assert(day >= 1 && day <= 31,
	      PyExc_ValueError,
	      "day must be in the range 1-31");
    Py_Assert(hour >= 0 && hour <= 23,
	      PyExc_ValueError,
	      "hour must be in the range 0-23");
    Py_Assert(minute >= 0 && minute <= 59,
	      PyExc_ValueError,
	      "minute must be in the range 0-59");
    Py_Assert((second >= 0.0 && second < 60.0) ||
	      (second == 60.0 && hour == 23 && minute == 59),
	      PyExc_ValueError,
	      "second must be in the range 0-<=60");
    
    time = newstruct(TIMESTAMP_STRUCT);
    if (!time) {
	PyErr_NoMemory();
	goto onError;
    }
    time->year = year;
    time->month = month;
    time->day = day;
    time->hour  = hour;
    time->minute = minute;
    time->second   = (UWORD)second;
    time->fraction = (UDWORD)((second - (double)time->second)
			      * FRACTIONS_IN_SECOND);

    var->data = (PTR)time;
    var->max_len = sizeof(TIMESTAMP_STRUCT);
    var->data_len = sizeof(TIMESTAMP_STRUCT);
    var->free_data = 1;
    var->ctype = SQL_C_TIMESTAMP;
    return 0;
    
 onError:
    return -1;
}

/* ------------------------------------------------------------------------ */

/* After successful execution of a SQL statement, this function sets
   up the array of output variables dbcs->vars for fetching the
   resulting data rows

   Returns -1 on error, 0 otherwise.

*/

static
int mxODBCursor_PrepareOutput(mxODBCursorObject *dbcs)
{
    int i;
    RETCODE rc;
    SWORD colcount;
    SDWORD rowcount;
    mxODBCursor_Variable *outvars = 0;

    /* Free previous output variables */
    mxODBCursor_FreeVars(dbcs);

    /* Get statistics */
    rc = SQLNumResultCols(dbcs->hstmt,&colcount);
    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
    dbcs->colcount = colcount;

    rc = SQLRowCount(dbcs->hstmt,&rowcount);
    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
    dbcs->rowcount = rowcount;

    DPRINTF("mxODBCursor_PrepareOutput: colcount=%i rowcount=%i\n",
	    colcount,rowcount);

    /* Prepare output variables */
    if (colcount > 0) {

	/* Allocate the variable array and clear it. */
	outvars = cnew(mxODBCursor_Variable,colcount);
	if (!outvars) {
	    PyErr_NoMemory();
	    goto onError;
	}

	/* Initialize the variables. */
	for (i = 0; i < colcount; i++) {
	    /* We store the precision as sqllen (the storage size in
	       bytes). This is only valid for character types. Numeric
	       types will have to determine the storage size based on
	       the used C type. */
	    rc = SQLDescribeCol(dbcs->hstmt,
				(UWORD)(i + 1),
				(UCHAR *)outvars[i].colname,
				(SWORD)sizeof(outvars[i].colname),
				&outvars[i].colname_len,
				&outvars[i].sqltype,
				&outvars[i].sqllen,
				&outvars[i].sqlscale,
				&outvars[i].nullable);
	    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
	    DPRINTF("mxODBCursor_PrepareOutput: column %i: "
		    "name='%s' type=%i precision=%i scale=%i nullable=%i\n",
		    i,outvars[i].colname,(int)outvars[i].sqltype,
		    (int)outvars[i].sqllen,(int)outvars[i].sqlscale,
		    (int)outvars[i].nullable);
	}
	dbcs->vars = outvars;

    }
    return 0;

 onError:
    if (outvars)
	free(outvars);
    return -1;
}

/* Allocate space for a single row fetch in dbcs->vars, set the
   C type and bind the variables if needed. 

   Returns -1 on error, 0 otherwise.

   Implementation notes:

   var->sqllen has to be set to the allocated size.  var->data_len
   should have the same value as var->sqllen; it will be overwritten
   by the true size during a fetch.

   The variable array is setup according to the var->sqltype values.
   It is only allocated once per result set.

*/

static
int mxODBCursor_AllocateOutputVars(mxODBCursorObject *dbcs)
{
    int i;
    SWORD sql_outparams = dbcs->colcount;
    mxODBCursor_Variable *vars = dbcs->vars;
    RETCODE rc;

    Py_Assert(sql_outparams > 0 && vars,
	      mxODBC_InterfaceError,
	      "Bad internal call to mxODBCursor_AllocateOutputVars");

    /* Allocation check */
    if (vars[0].data)
	return 0;

    /* Unbind any old variables that might still be bound */
    rc = SQLFreeStmt(dbcs->hstmt,SQL_UNBIND);
    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    /* Allocate a new set of variables and bind them to the columns */
    for (i = 0; i < sql_outparams; i++) {
	PTR data = vars[i].data;

	/* Sanity check. */
	Py_Assert(data == NULL,
		  mxODBC_InterfaceError,
		  "Memory leakage in mxODBCursor_AllocateOutputVars");

	/* Special case: the exact size is unkown, so reserve some
	   standard amount of bytes and then let mxODBCursor_GetData()
	   handle the fetching... in some cases we can reset this flag
	   because we can determine the size ourselves (see below).  */
	if (vars[i].sqllen == 0) {
	    vars[i].sqllen = DEFAULT_FIELD_SIZE;
	    vars[i].use_getdata = 1;
	}
	else
	    vars[i].use_getdata = 0;
	
	switch (vars[i].sqltype) {

	case SQL_CHAR:
	case SQL_VARCHAR:
	    /* 0-terminated, so we need to reserve one more byte */
	    vars[i].sqllen = vars[i].sqllen + 1;
	    data = (PTR)new(char,vars[i].sqllen);
	    if (!data) {
		PyErr_NoMemory();
		goto onError;
	    }
	    vars[i].data = data;
	    vars[i].free_data = 1;
	    vars[i].ctype = SQL_C_CHAR;
	    vars[i].convert = &mxODBC_StringFromString;
	    break;
		
	case SQL_LONGVARCHAR:
	    data = (PTR)new(char,DEFAULT_LONG_SIZE);
	    if (!data) {
		PyErr_NoMemory();
		goto onError;
	    }
	    vars[i].data = data;
	    vars[i].sqllen = DEFAULT_LONG_SIZE;
	    vars[i].free_data = 1;
	    vars[i].use_getdata = 1;
#ifdef ADABAS_WORKAROUND
	    /* We use BINARY here because of an Adabas bug */
      	    vars[i].ctype = SQL_C_BINARY;
#else
      	    vars[i].ctype = SQL_C_CHAR;
#endif
	    vars[i].convert = &mxODBC_StringFromString;
	    break;

	case SQL_BINARY:
	case SQL_VARBINARY:
	    data = (PTR)new(char,vars[i].sqllen);
	    if (!data) {
		PyErr_NoMemory();
		goto onError;
	    }
	    vars[i].data = data;
	    vars[i].free_data = 1;
	    vars[i].ctype = SQL_C_BINARY;
	    vars[i].convert = &mxODBC_StringFromString;
	    break;

	case SQL_LONGVARBINARY:
	    data = (PTR)new(char,DEFAULT_LONG_SIZE);
	    if (!data) {
		PyErr_NoMemory();
		goto onError;
	    }
	    vars[i].data = data;
	    vars[i].sqllen = DEFAULT_LONG_SIZE;
	    vars[i].free_data = 1;
	    vars[i].use_getdata = 1;
	    vars[i].ctype = SQL_C_BINARY;
	    vars[i].convert = &mxODBC_StringFromString;
	    break;

	case SQL_DATE:
	    if (dbcs->datetimeformat == MXODBC_DATETIME_DATETIME)
		vars[i].convert = &mxODBC_DateTimeFromDateStruct;
	    else if (dbcs->datetimeformat == MXODBC_TUPLE_DATETIME)
		vars[i].convert = &mxODBC_TupleFromDateStruct;
	    else if (dbcs->datetimeformat == MXODBC_FLOAT_DATETIME)
		vars[i].convert = &mxODBC_FloatFromDateStruct;
	    else { 
		/* Use a string; formatting is done by the ODBC driver */
		vars[i].sqllen = 40;
		data = (PTR)new(char,vars[i].sqllen);
		if (!data) {
		    PyErr_NoMemory();
		    goto onError;
		}
		vars[i].data = data;
		vars[i].free_data = 1;
		vars[i].ctype = SQL_C_CHAR;
		vars[i].convert = &mxODBC_StringFromString;
		break;
	    }
	    data = (PTR)new(char,sizeof(DATE_STRUCT));
	    if (!data) {
		PyErr_NoMemory();
		goto onError;
	    }
	    vars[i].data = data;
	    vars[i].sqllen = sizeof(DATE_STRUCT);
	    vars[i].free_data = 1;
	    vars[i].ctype = SQL_C_DATE;
	    break;
	    
	case SQL_TIME:
	    if (dbcs->datetimeformat == MXODBC_DATETIME_DATETIME)
		vars[i].convert = &mxODBC_DateTimeDeltaFromTimeStruct;
	    else if (dbcs->datetimeformat == MXODBC_TUPLE_DATETIME)
		vars[i].convert = &mxODBC_TupleFromTimeStruct;
	    else if (dbcs->datetimeformat == MXODBC_FLOAT_DATETIME)
		vars[i].convert = &mxODBC_FloatFromTimeStruct;
	    else { 
		/* Use a string; formatting is done by the ODBC driver */
		vars[i].sqllen = 40;
		data = (PTR)new(char,vars[i].sqllen);
		if (!data) {
		    PyErr_NoMemory();
		    goto onError;
		}
		vars[i].data = data;
		vars[i].free_data = 1;
		vars[i].ctype = SQL_C_CHAR;
		vars[i].convert = &mxODBC_StringFromString;
		break;
	    }
	    data = (PTR)new(char,sizeof(TIME_STRUCT));
	    if (!data) {
		PyErr_NoMemory();
		goto onError;
	    }
	    vars[i].data = data;
	    vars[i].sqllen = sizeof(TIME_STRUCT);
	    vars[i].free_data = 1;
	    vars[i].ctype = SQL_C_TIME;
	    break;
	    
	case SQL_TIMESTAMP:
	    if (dbcs->datetimeformat == MXODBC_DATETIME_DATETIME)
		vars[i].convert = &mxODBC_DateTimeFromTimestampStruct;
	    else if (dbcs->datetimeformat == MXODBC_TUPLE_DATETIME)
		vars[i].convert = &mxODBC_TupleFromTimestampStruct;
	    else if (dbcs->datetimeformat == MXODBC_FLOAT_DATETIME)
		vars[i].convert = &mxODBC_FloatFromTimestampStruct;
	    else { 
		/* Use a string; formatting is done by the ODBC driver */
		vars[i].sqllen = 40;
		data = (PTR)new(char,vars[i].sqllen);
		if (!data) {
		    PyErr_NoMemory();
		    goto onError;
		}
		vars[i].data = data;
		vars[i].free_data = 1;
		vars[i].ctype = SQL_C_CHAR;
		vars[i].convert = &mxODBC_StringFromString;
		break;
	    }
	    data = (PTR)new(char,sizeof(TIMESTAMP_STRUCT));
	    if (!data) {
		PyErr_NoMemory();
		goto onError;
	    }
	    vars[i].data = data;
	    vars[i].sqllen = sizeof(TIMESTAMP_STRUCT);
	    vars[i].free_data = 1;
	    vars[i].ctype = SQL_C_TIMESTAMP;
	    break;
	    
	case SQL_TINYINT:
	case SQL_SMALLINT:
	case SQL_INTEGER:
	case SQL_BIT:
	    /* For numeric data the value of sqllen is unusable since
	       it points to the precision of the type (number of
	       decimals) and not the storage size. */
	    vars[i].sqllen = sizeof(long);
	    data = (PTR)new(char,vars[i].sqllen);
	    if (!data) {
		PyErr_NoMemory();
		goto onError;
	    }
	    vars[i].data = data;
	    vars[i].free_data = 1;
	    vars[i].ctype = SQL_C_SLONG;
	    vars[i].convert = &mxODBC_IntFromLong;
	    break;

	case SQL_BIGINT:
	    /* We convert BIGINT values to Python Long objects and
	       use strings as intermediate format for this. Since
	       integers have no scale, the precision stored in sqllen
	       gives a correct hint at the size of the needed
	       buffer. Strings are 0-terminated, so we need to reserve
	       one more byte for the buffer.

	       Note that according to the ODBC docs, BIGINT values fit
	       into 64 bits, so a long long could also do the trick in
	       some future version of mxODBC.

	    */
	    vars[i].sqllen = vars[i].sqllen + 1;
	    data = (PTR)new(char,vars[i].sqllen);
	    if (!data) {
		PyErr_NoMemory();
		goto onError;
	    }
	    vars[i].data = data;
	    vars[i].free_data = 1;
	    vars[i].ctype = SQL_C_CHAR;
	    vars[i].convert = &mxODBC_LongFromString;
	    break;
		
	case SQL_DECIMAL:
	case SQL_NUMERIC:
	case SQL_REAL:
	case SQL_FLOAT:
	case SQL_DOUBLE:
	    /* For numeric data the value of sqllen is unusable since
	       it points to the precision of the type (number of
	       decimals) and not the storage size. */
	    vars[i].sqllen = sizeof(double);
	    data = (PTR)new(char,vars[i].sqllen);
	    if (!data) {
		PyErr_NoMemory();
		goto onError;
	    }
	    vars[i].data = data;
	    vars[i].free_data = 1;
	    vars[i].ctype = SQL_C_DOUBLE;
	    vars[i].convert = &mxODBC_FloatFromDouble;
	    break;

	default:
	    Py_ErrorWithArg(mxODBC_InterfaceError,
			    "SQL type (code %i) not implemented",
			    vars[i].sqltype);
	    break;
	}

	DPRINTF("mxODBCursor_AllocateOutputVars: "
		"binding column %i - sqltype=%i, ctype=%i, free=%i, maxlen=%i\n",
		i,(int)vars[i].sqltype,(int)vars[i].ctype,
		(int)vars[i].free_data,(int)vars[i].sqllen);

	/* Bind the column */
	if (!vars[i].use_getdata) {

	    /* vars[i].data_len will be overwritten by the true length
	       of the data available or set to SQL_NO_TOTAL. The docs
	       say it's an input variable (?!), so we assign the
	       maximum length here just to make sure. */
	    vars[i].data_len = vars[i].sqllen;
	    rc = SQLBindCol(dbcs->hstmt,
			    (UWORD)(i + 1),
			    vars[i].ctype,
			    vars[i].data,
			    vars[i].sqllen,	/* max size */
			    &vars[i].data_len);	/* true size */
	    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
	    DPRINTF("mxODBCursor_AllocateOutputVars:   true len=%i\n",
		    (int)vars[i].data_len);

	    /* Guard against memory corruption... */
	    Py_Assert(vars[i].data_len <= (SDWORD)vars[i].sqllen,
		      mxODBC_InterfaceError,
		      "Internal error in mxODBCursor_AllocateOutputVars");

	}
	else {
	    /* Unbind the column... just in case */
	    rc = SQLBindCol(dbcs->hstmt,
			    (UWORD)(i + 1),
			    vars[i].ctype,
			    NULL,
			    0,
			    NULL);
	    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
	}
    }

    return 0;

 onError:
    /* Unbind all variables already bound */
    rc = SQLFreeStmt(dbcs->hstmt,SQL_UNBIND);
    IF_DEBUGGING {
	if (rc != SQL_SUCCESS)
	    DPRINTF("mxODBCursor_AllocateOutputVars: "
		    "variable unbind returned an error: rc=%i",
		    rc);
    }
    /* Release reserved storage and reset */
    for (i = 0; i < sql_outparams; i++) {
	if (vars[i].free_data) {
	    if (vars[i].data)
		free(vars[i].data);
	    vars[i].free_data = 0;
	}
	vars[i].data = 0;
	vars[i].data_len = 0;
    }
    return -1;
}

/* Get a column from the last fetch using SQLGetData; allocate more
   room in var[column].data as necessary 
   - returns -1 on error, 0 otherwise
*/

static 
int mxODBCursor_GetData(mxODBCursorObject *dbcs,
			 int column)
{
    int offset = 0;
    mxODBCursor_Variable *var = &dbcs->vars[column];
    int null_terminated = (var->ctype == SQL_C_CHAR);
    RETCODE rc;

    Py_Assert(var->data != NULL,
	      mxODBC_InterfaceError,
	      "internal error in mxODBCursor_GetData: data buffer is NULL");

    var->data_len = 0;

    /* XXX There is a problem with LONGVARCHAR fields under Adabas D 6.1.1:
       XXX byte position 32766 is always 0 in data longer than 32768
       XXX bytes when written as SQL_C_CHAR. We work around this bug by using
       XXX SQL_C_BINARY -- probably costs a little performance though
       XXX because of the additional conversion.

       Note: This problem does not occur with LONGVARBINARY fields. 
    */
    while (1) {
	SDWORD received;
	
	Py_BEGIN_ALLOW_THREADS;
	rc = SQLGetData(dbcs->hstmt,
			(UWORD)(column + 1),
			var->ctype,
			(PTR)((char *)var->data + offset),
			var->sqllen - offset,
			&received);
	Py_END_ALLOW_THREADS;

	/* NULL check */
	if (received == SQL_NULL_DATA) {
	    var->data_len = received;
	    break;
	}

	DPRINTF("mxODBCursor_GetData: received %i (max. %i) bytes, "
		"buf size = %i, offset = %i, ctype = %i\n",\
		(int)received,(int)(var->sqllen - offset),
		(int)var->sqllen,(int)offset,(int)var->ctype);

	/* Finished ? */
	if (rc == SQL_SUCCESS) {
	    var->data_len += received;
	    break;
	}

	/* There's more, so resize the data area */
	else if (rc == SQL_SUCCESS_WITH_INFO) {
	    int newsize;
	    void *newdata;
	    newsize = var->sqllen;
	    if (newsize < 32768)
		newsize = var->sqllen * 2; /* grow 100% */
	    else
		newsize += newsize >> 1; /* grow 50% */
	    newdata = (void*)resize(var->data,char,newsize);
	    if (!newdata) {
		PyErr_NoMemory();
		goto onError;
	    }
	    received = var->sqllen - offset;
	    /* Compensate for the \0-byte */
	    if (null_terminated)
		received--;
	    var->data_len += received;
	    offset += received;
	    /* Now switch to the resized buffer */
	    var->data = newdata;
	    var->sqllen = newsize;
	}

	/* Hmm, now what ? */
	else
	    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
    }
#if 0
    /* Check for the Adabas bug */
    if (null_terminated && var->data_len > 32766 && 
	((char*)var->data)[32766] == '\0')
	Py_Error(mxODBC_InterfaceError,
		 "ADABAS LONGVARCHAR error: data unusable !");
#endif
    return 0;
    
 onError:
    return -1;
}


/* Fetch a single row from the cursor dbcs and return it as tuple.
   On error, NULL is returned.
   - The function allocates space for the fetch as needed.
*/

static
PyObject *mxODBCursor_Fetch(mxODBCursorObject *dbcs)
{
    PyObject *w,*v = 0;
    int i;
    SWORD sql_outparams = dbcs->colcount;
    mxODBCursor_Variable *vars = dbcs->vars;
    RETCODE rc;

    Py_Assert(sql_outparams > 0,
	      mxODBC_ProgError,
	      "missing result set");

    Py_Assert(vars != NULL,
	      mxODBC_InterfaceError,
	      "internal error: mxODBCursor_fetch called, but vars not set");

    /* Allocate & bind output variables if not already done */
    if (!vars[0].data)
	if (mxODBCursor_AllocateOutputVars(dbcs))
	    goto onError;
    
    /* Fetch */
    DPRINTF("mxODBCursor_Fetch: fetching %i column(s).\n",(int)sql_outparams);

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLFetch(dbcs->hstmt);
    Py_END_ALLOW_THREADS;

    if (rc == SQL_NO_DATA_FOUND)
	Py_ReturnNone();
    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    /* Create output */
    v = PyTuple_New(sql_outparams);
    if (!v)
	goto onError;
    
    for (i = 0; i < sql_outparams; i++) {
	DPRINTF("mxODBCursor_Fetch:  column %i...\n",i);
	if (vars[i].use_getdata)
	    if (mxODBCursor_GetData(dbcs,i))
		goto onError;
	if (vars[i].data_len == SQL_NULL_DATA) {
	    Py_INCREF(Py_None);
	    w = Py_None;
	}
	else {
	    w = vars[i].convert(&vars[i]);
	    if (!w)
		goto onError;
	}
	Py_PRINT_REFCOUNT(w);
	PyTuple_SET_ITEM(v,i,w);
    }
    
    return v;

 onError:
    Py_XDECREF(v);
    return NULL;
}

/* Fetch at most nrows rows from the cursor dbcs and return it as a
   list of tuples. On error, NULL is returned.

   The function allocates space for the fetch as needed.

   XXX Speedup by using arrays of rows and SQLParamOptions().
*/

static
PyObject *mxODBCursor_FetchMany(mxODBCursorObject *dbcs,
			       int nrows)
{
    PyObject *w,*v = 0;
    PyObject *list = 0;
    int i,j;
    SWORD sql_outparams = dbcs->colcount;
    mxODBCursor_Variable *vars = dbcs->vars;
    RETCODE rc;

    Py_Assert(sql_outparams > 0,
	      mxODBC_ProgError,
	      "missing result set");

    Py_Assert(vars != NULL,
	      mxODBC_InterfaceError,
	      "internal error: mxODBCursor_fetch called, but vars not set");

    /* Allocate & bind output variables if not already done */
    if (!vars[0].data)
	if (mxODBCursor_AllocateOutputVars(dbcs))
	    goto onError;
    
    /* Create output */
    list = PyList_New(nrows);
    if (!list) 
	goto onError;
    
    DPRINTF("mxODBCursor_FetchMany: fetching %i column(s) and "
	    "at most %i row(s).\n",
	    (int)sql_outparams,nrows);

    for (j = 0; j < nrows; j++) {
    
	DPRINTF("mxODBCursor_FetchMany:   row %i...\n",j);

	/* Fetch */

	Py_BEGIN_ALLOW_THREADS;
	rc = SQLFetch(dbcs->hstmt);
	Py_END_ALLOW_THREADS;
	    
	if (rc == SQL_NO_DATA_FOUND)
	    break;
	mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

	v = PyTuple_New(sql_outparams);
	if (!v)
	    goto onError;
    
	for (i = 0; i < sql_outparams; i++) {
	    if (vars[i].use_getdata)
		if (mxODBCursor_GetData(dbcs,i))
		    goto onError;
	    if (vars[i].data_len == SQL_NULL_DATA) {
		Py_INCREF(Py_None);
		w = Py_None;
	    }
	    else {
		w = vars[i].convert(&vars[i]);
		if (!w)
		    goto onError;
	    }
	    Py_PRINT_REFCOUNT(w);
	    PyTuple_SET_ITEM(v,i,w);
	}

	Py_PRINT_REFCOUNT(v);
	PyList_SET_ITEM(list,j,v);
	v = 0;
    }
    
    DPRINTF("mxODBCursor_FetchMany: done -- read %i row(s).\n",j);

    /* If the fetch didn't result in nrows, delete the remaining
       entries in list */
    if (j < nrows)
	PyList_SetSlice(list,j,nrows,(PyObject*)NULL);

    return list;

 onError:
    Py_XDECREF(v);
    Py_XDECREF(list);
    return NULL;
}

/* Fetch all rows from the cursor dbcs and return it as a
   list of tuples. On error, NULL is returned.

   The function allocates space for the fetch as needed.

   XXX Speedup by using arrays of rows and SQLParamOptions().
*/

static
PyObject *mxODBCursor_FetchAll(mxODBCursorObject *dbcs)
{
    PyObject *w,*v = 0;
    PyObject *list = 0;
    int i;
    SWORD sql_outparams = dbcs->colcount;
    mxODBCursor_Variable *vars = dbcs->vars;
    RETCODE rc;

    Py_Assert(sql_outparams > 0,
	      mxODBC_ProgError,
	      "missing result set");

    Py_Assert(vars != NULL,
	      mxODBC_InterfaceError,
	      "internal error: mxODBCursor_fetch called, but vars not set");

    /* Allocate & bind output variables if not already done */
    if (!vars[0].data)
	if (mxODBCursor_AllocateOutputVars(dbcs))
	    goto onError;

    /* Create output */
    list = PyList_New(0);
    if (!list)
	goto onError;
    
    DPRINTF("mxODBCursor_FetchAll: fetching %i column(s).\n",
	    (int)sql_outparams);

    while (1) {
    
	DPRINTF("mxODBCursor_FetchAll:   row %i...\n",
		PyList_GET_SIZE(list));

	/* Fetch one row */

	Py_BEGIN_ALLOW_THREADS;
	rc = SQLFetch(dbcs->hstmt);
	Py_END_ALLOW_THREADS;

	if (rc == SQL_NO_DATA_FOUND)
	    break;
	mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

	v = PyTuple_New(sql_outparams);
	if (!v)
	    goto onError;
    
	for (i = 0; i < sql_outparams; i++) {
	    if (vars[i].use_getdata)
		if (mxODBCursor_GetData(dbcs,i))
		    goto onError;
	    if (vars[i].data_len == SQL_NULL_DATA) {
		Py_INCREF(Py_None);
		w = Py_None;
	    }
	    else {
		w = vars[i].convert(&vars[i]);
		if (!w)
		    goto onError;
	    }
	    Py_PRINT_REFCOUNT(w);
	    PyTuple_SET_ITEM(v,i,w);
	}

	PyList_Append(list,v);
	Py_DECREF(v);
	v = 0;
    }
    
    DPRINTF("mxODBCursor_FetchAll: done -- read %i row(s).\n",
	    PyList_GET_SIZE(list));

    return list;

 onError:
    Py_XDECREF(v);
    Py_XDECREF(list);
    return NULL;
}

/* Bind the data from the Python input variable v to the cursor
   variable var. i and j indicate the column and row of the variable
   in the input field. j is only used to produce verbose error
   messages. i is needed for the SQL API. 

   XXX Should try to reuse any previously allocated memory.

*/

static 
int mxODBCursor_BindInputParameter(mxODBCursorObject *dbcs,
				   register mxODBCursor_Variable *var,
				   register PyObject *v,
				   int i,
				   int j)
{
    PTR old_data;
    SWORD old_ctype;
    SDWORD old_data_len;

    /* Free any previously allocated variable buffer and reset the
       data structure */
    if (var->free_data && var->data) {
	free(var->data);
	var->free_data = 0;
	var->data = 0;
	var->data_len = 0;
	var->max_len = 0;
    }
    else {
	/* If max_len is 0 and data is non-NULL, the max_len values
	   will be set to the same value as data_len below. */
	var->max_len = 0;
    }

    /* Remember the old values to prevent rebinding to the
       same data. */
    old_data = var->data;
    old_data_len = var->data_len;
    old_ctype = var->ctype;

    /* Prepare the binding.

       Three modes are supported:

       1. NULL binding (Python's None is used for this)
       2. SQL type binding (the database is queried for the data type
          to send)
       3. Python type binding (the Python type decides which data type
          is sent)

       XXX Always make sure that the following variables are set since these
           are not reset by the code above:

	   var->sqltype
	   var->sqllen
	   var->sqlscale
	   var->ctype

    */

    if (v == Py_None) {
	/* Let None emulate the SQL NULL value; XXX Should check
           var->nullable... oh well, the database will issue an error
           anyway. */
	DPRINTF("mxODBCursor_BindInputParameter: "
		"binding row %i, column %i to NULL\n",
		j,i);
	var->data = 0;
	var->data_len = SQL_NULL_DATA;
	var->sqltype = SQL_CHAR;
	var->sqllen = 0;
	var->sqlscale = 0;
	var->max_len = 0;
	var->ctype = SQL_C_CHAR;
    }

    /* Choose which binding method to use. A var->sqltype
       SQL_TYPE_NULL forces usage of the Python binding method since
       in that case no useable SQL type information is available. */

    else if (dbcs->dbc->bindmethod == MXODBC_BIND_USING_SQLTYPE &&
	     var->sqltype != SQL_TYPE_NULL) {

	/* Convert variable data to a compatible SQL type. 

	   The approach taken here asks the driver/manager which type
	   it wants and then does proper conversion in Python rather
	   than letting the driver/manager take care of it.

	   Note that this section assumes the var->sqltype,
	   var->sqllen and var->sqlscale to be already properly set
	   up.

	   XXX The following assumes that the ODBC driver will
	   XXX only read the passed in data (and not write to it) !!!

	*/
	DPRINTF("mxODBCursor_BindInputParameter: "
		"binding row %i, column %i using the SQL type (%i)\n",
		j,i,var->sqltype);

	switch (var->sqltype) {

	case SQL_CHAR:
	case SQL_VARCHAR:
#ifndef ADABAS_WORKAROUND
	case SQL_LONGVARCHAR:
#endif
	    if (PyString_Check(v)) {
		var->data = (PTR)(PyString_AS_STRING(v));
		var->data_len = PyString_GET_SIZE(v);
		var->ctype = SQL_C_CHAR;
	    }
	    else {
		/* Hmm, maybe we can turn it into a string
		   then... */
		if (mxODBC_StringFromObject(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	      "argument %i in tuple %i: string conversion error",
				     i,j);
		    goto onError;
		}
	    }
	    break;

#ifdef ADABAS_WORKAROUND
	case SQL_LONGVARCHAR:
#endif
	case SQL_BINARY:
	case SQL_VARBINARY:
	case SQL_LONGVARBINARY:
	    if (PyString_Check(v)) {
		var->data = (PTR)(PyString_AS_STRING(v));
		var->data_len = PyString_GET_SIZE(v);
		var->ctype = SQL_C_BINARY;
	    }
	    else {
		/* Hmm, maybe we can turn it into a string
		   then... */
		if (mxODBC_BinaryFromObject(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	      "argument %i in tuple %i: binary conversion error",
				     i,j);
		    goto onError;
		}
	    }
	    break;

	case SQL_TINYINT:
	case SQL_SMALLINT:
	case SQL_INTEGER:
	case SQL_BIT:
	    if (PyInt_Check(v)) {
		var->data = (PTR)&(PyInt_AS_LONG(v));
		var->data_len = sizeof(long);
		var->ctype = SQL_C_SLONG;
	    }
	    else {
		/* Hmm, maybe we can turn it into a C long
		   then... */
		if (mxODBC_LongFromObject(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	      "argument %i in tuple %i: integer conversion error",
				     i,j);
		    goto onError;
		}
	    }
	    break;

	case SQL_BIGINT:
	    if (PyLong_Check(v)) {
		if (mxODBC_StringFromLongObject(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	      "argument %i in tuple %i: long conversion error",
				     i,j);
		    goto onError;
		}
	    }
	    else {
		/* Hmm, maybe we can use the string
		   representation then... */
		if (mxODBC_StringFromObject(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	      "argument %i in tuple %i: long conversion error",
				     i,j);
		    goto onError;
		}
	    }
	    break;

	case SQL_DECIMAL:
	case SQL_NUMERIC:
	case SQL_REAL:
	case SQL_FLOAT:
	case SQL_DOUBLE:
	    if (PyFloat_Check(v)) {
		var->data = (PTR)&(PyFloat_AS_DOUBLE(v));
		var->data_len = sizeof(double);
		var->ctype = SQL_C_DOUBLE;
	    }
	    else {
		/* Hmm, maybe we can turn it into a double
		   then... */
		if (mxODBC_DoubleFromObject(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	      "argument %i in tuple %i: float conversion error",
				     i,j);
		    goto onError;
		}
	    }
	    break;

	case SQL_DATE: 
	    if (mxDateTime_Check(v)) {
		if (mxODBC_DateStructFromDateTime(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	   "argument %i in tuple %i: DateTime conversion error",
				     i,j);
		    goto onError;
		}
		break;
	    }
	    else if (PyTuple_Check(v)) {
		if (mxODBC_DateStructFromTuple(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	   "argument %i in tuple %i must be a mxODBC date tuple",
				     i,j);
		    goto onError;
		}
		break;
	    }
	    else if (PyNumber_Check(v)) {
		if (mxODBC_DateStructFromFloat(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	   "argument %i in tuple %i must be a mxODBC date float",
				     i,j);
		    goto onError;
		}
	    }
	    else if (PyString_Check(v)) {
		var->data = (PTR)(PyString_AS_STRING(v));
		var->data_len = strlen(PyString_AS_STRING(v));
		var->ctype = SQL_C_CHAR;
	    }
	    else {
		PyErr_Format(PyExc_TypeError,
		"argument %i in tuple %i must be a date value",
			     i,j);
		goto onError;
	    }
	    break;

	case SQL_TIME: 
	    if (mxDateTimeDelta_Check(v)) {
		if (mxODBC_TimeStructFromDateTimeDelta(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
      "argument %i in tuple %i: DateTimeDelta conversion error",
				     i,j);
		    goto onError;
		}
		break;
	    }
	    else if (PyTuple_Check(v)) {
		if (mxODBC_TimeStructFromTuple(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	   "argument %i in tuple %i must be a mxODBC time tuple",
				     i,j);
		    goto onError;
		}
		break;
	    }
	    else if (PyNumber_Check(v)) {
		if (mxODBC_TimeStructFromFloat(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	   "argument %i in tuple %i must be a mxODBC time float",
				     i,j);
		    goto onError;
		}
	    }
	    else if (PyString_Check(v)) {
		var->data = (PTR)(PyString_AS_STRING(v));
		var->data_len = strlen(PyString_AS_STRING(v));
		var->ctype = SQL_C_CHAR;
	    }
	    else {
		PyErr_Format(PyExc_TypeError,
		"argument %i in tuple %i must be a time value",
			     i,j);
		goto onError;
	    }
	    break;

	case SQL_TIMESTAMP:
	    if (mxDateTime_Check(v)) {
		if (mxODBC_TimestampStructFromDateTime(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
	   "argument %i in tuple %i: DateTime conversion error",
				     i,j);
		    goto onError;
		}
		break;
	    }
	    else if (PyTuple_Check(v)) {
		if (mxODBC_TimestampStructFromTuple(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
      "argument %i in tuple %i must be a mxODBC timestamp tuple",
				     i,j);
		    goto onError;
		}
		break;
	    }
	    else if (PyNumber_Check(v)) {
		if (mxODBC_TimestampStructFromFloat(v,var)) {
		    if (!PyErr_Occurred())
			PyErr_Format(PyExc_TypeError,
      "argument %i in tuple %i must be a mxODBC timestamp float",
				     i,j);
		    goto onError;
		}
	    }
	    else if (PyString_Check(v)) {
		var->data = (PTR)(PyString_AS_STRING(v));
		var->data_len = strlen(PyString_AS_STRING(v));
		var->ctype = SQL_C_CHAR;
	    }
	    else {
		PyErr_Format(PyExc_TypeError,
		"argument %i in tuple %i must be a timestamp value",
			     i,j);
		goto onError;
	    }
	    break;

	default:
	    Py_ErrorWithArg(mxODBC_InterfaceError,
			    "SQL type %i not implemented",
			    var->sqltype);
	    break;
	}
    }

    else if (dbcs->dbc->bindmethod == MXODBC_BIND_USING_PYTHONTYPE ||
	     var->sqltype == SQL_TYPE_NULL) {

	/* Convert variable data to a compatible SQL type. 

	   This approach looks at the Python object type and tries to
	   figure out which data type to choose as basis for
	   communication with the driver/manager.

	   Note that since no column information is available, we have
	   to set var->sqllen and var->sqlscale in addition to
	   var->sqltype according to the values we pass to the
	   SQLBindParameter() API. See the ODBC Doc Appendix on scales
	   and precision for details.

	   XXX The following assumes that the ODBC driver will
	   XXX only read the passed in data (and not write to it) !!!

	*/
	DPRINTF("mxODBCursor_BindInputParameter: "
		"binding row %i, column %i using the Python type (%s)\n",
		j,i,v->ob_type->tp_name);

	/* XXX For instances add a way for them to choose how they
	       would like to be bound. */
	
	if (PyString_Check(v)) {
	    var->data = (PTR)(PyString_AS_STRING(v));
	    var->data_len = PyString_GET_SIZE(v);
	    if (var->data_len < 256)
		var->sqltype = SQL_CHAR;
	    else
		var->sqltype = SQL_LONGVARCHAR;
#ifdef ADABAS_WORKAROUND
	    var->ctype = SQL_C_BINARY;
#else
	    var->ctype = SQL_C_CHAR;
#endif
	    var->sqllen = var->data_len;
	    var->sqlscale = 0;
	}

	else if (PyInt_Check(v)) {
	    var->data = (PTR)&(PyInt_AS_LONG(v));
	    var->data_len = sizeof(long);
	    var->sqltype = SQL_INTEGER;
	    var->ctype = SQL_C_SLONG;
	    var->sqllen = 0;
	    var->sqlscale = 0;
	}

	else if (PyLong_Check(v)) {
	    if (mxODBC_StringFromLongObject(v,var)) {
		if (!PyErr_Occurred())
		    PyErr_Format(PyExc_TypeError,
		    "argument %i in tuple %i: long conversion error",
				 i,j);
		goto onError;
	    }
	    var->sqltype = SQL_CHAR;
	    var->sqllen = var->data_len;
	    var->sqlscale = 0;
	}

	else if (PyFloat_Check(v)) {
	    var->data = (PTR)&(PyFloat_AS_DOUBLE(v));
	    var->data_len = sizeof(double);
	    var->sqltype = SQL_DOUBLE;
	    var->ctype = SQL_C_DOUBLE;
	    var->sqllen = 0;
	    var->sqlscale = 0;
	}

	else if (mxDateTime_Check(v)) {
	    if (mxODBC_TimestampStructFromDateTime(v,var)) {
		if (!PyErr_Occurred())
		    PyErr_Format(PyExc_TypeError,
		    "argument %i in tuple %i: DateTime conversion error",
				 i,j);
		goto onError;
	    }
	    var->sqltype = SQL_TIMESTAMP;
	    /* Not sure whether these are needed since we pass in a
	       TIMESTAMP_STRUCT and not a string; sqllen is the number
	       of characters in the used string format and sqlscale
	       the number of decimal points used in the seconds
	       part. */
	    var->sqllen = 22;
	    var->sqlscale = 2; 
	}

	else if (mxDateTimeDelta_Check(v)) {
	    if (mxODBC_TimeStructFromDateTimeDelta(v,var)) {
		if (!PyErr_Occurred())
		    PyErr_Format(PyExc_TypeError,
		    "argument %i in tuple %i: DateTimeDelta conversion error",
				 i,j);
		goto onError;
	    }
	    var->sqltype = SQL_TIME;
	    var->sqllen = 0;
	    var->sqlscale = 0;
	}

	else {
	    /* Hmm, maybe we can turn it into a string then... */
	    if (mxODBC_StringFromObject(v,var)) {
		if (!PyErr_Occurred())
		    PyErr_Format(PyExc_TypeError,
		    "argument %i in tuple %i: string conversion error",
				 i,j);
		goto onError;
	    }
	    var->sqltype = SQL_CHAR;
	    var->sqllen = var->data_len;
	    var->sqlscale = 0;
	}
    }

    else
	Py_ErrorWithArg(mxODBC_InterfaceError,
			"failed to bind input variable (binding method: %i)",
			dbcs->dbc->bindmethod);

    /* For input variables that bind directly to Python objects
       instead of allocated buffers, max_len is equal to data_len. For
       NULL data pointers, the maximum buffer length is set to 0 (MS
       ODBC Manager needs this).  */
    if (var->data) {
	if (var->max_len == 0)
	    var->max_len = var->data_len;
    }
    else
	var->max_len = 0;

    if (DEBUGGING) {
	DPRINTF("mxODBCursor_BindInputParameter: binding row %i, column %i "
		"- sqltype=%i, ctype=%i, sqllen=%i, data_len=%i, max_len=%i\n",
		j,i,(int)var->sqltype,(int)var->ctype,
		(int)var->sqllen,(int)var->data_len,(int)var->max_len);
	if (var->ctype == SQL_C_CHAR &&
	    var->data != NULL) {
	    DPRINTF("mxODBCursor_BindInputParameter:  "
		    "is string data [:256] = '%.*s'\n",
		    min((int)var->data_len,256),
		    (char*)var->data);
	}
    }

    /* Performance shortcut: don't bind to the same buffer twice. */
    if (var->data == old_data &&
	var->ctype == old_ctype &&
	var->data_len == old_data_len) {
	DPRINTF("mxODBCursor_BindInputParameter:  "
		"skipping the API call - same buffer\n");
    }
    else {
	RETCODE rc;

	/* Bind the parameter to the variable */
	rc = SQLBindParameter(dbcs->hstmt,
			      (UWORD)(i + 1),
			      SQL_PARAM_INPUT,
			      var->ctype,
			      var->sqltype,
			      var->sqllen,
			      var->sqlscale,
			      var->data,
			      var->max_len,
			      &var->data_len);
	mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

	DPRINTF("mxODBCursor_BindInputParameter: bound data length=%i\n",
		(int)var->data_len);
    }

    return 0;
    
 onError:
    return -1;
}


/* Execute the SQL command in sqlcmd using the parameters in params.
   The function returns 0 on sucess, -1 otherwise, and sets the
   appropriate values in dbcs.
   - params may be NULL if no parameters are needed
   - params can be either a list of tuples or a single tuple
   - sqlcmd may be NULL; in that case the prepare step is omitted,
     so the last prepared statement is used again
   - After successful completion, the function sets up the dbcs->vars
     array to match the result columns using mxODBCursor_PrepareOutput()
   - WARNING: call mxODBCursor_Reset() before doing implicit executes
     (like with SQLTables)
   - returns -1 on error, 0 otherwise
*/

static
int mxODBCursor_Execute(mxODBCursorObject *dbcs,
			char *sqlcmd,
			int sqlcmd_len,
			PyObject *params)
{
    PyObject *w;
    int i,j,nrows = 0;
    SWORD sql_inparams;
    mxODBCursor_Variable *vars = 0;
    RETCODE rc;

    if (params) {
	if (PyList_Check(params)) {
	    nrows = PyList_GET_SIZE(params);
	    /* If no tuples are given, ignore the params list */
	    if (nrows == 0)
		params = 0;
	}
	else if (PyTuple_Check(params))
	    nrows = 1;
	else
	    Py_Error(PyExc_TypeError,
		     "optional second argument must be a list or tuple");
    }

#if 0
    /* Unbind any old parameters that might still be bound */
    rc = SQLFreeStmt(dbcs->hstmt,SQL_RESET_PARAMS);
    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    /* Unbind any old variables that might still be bound */
    rc = SQLFreeStmt(dbcs->hstmt,SQL_UNBIND);
    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
#endif

    /* Close a possibly open statement cursor */
    rc = SQLFreeStmt(dbcs->hstmt,SQL_CLOSE);
    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    /* Prepare command */
    if (sqlcmd) {
	DPRINTF("mxODBCursor_Execute: preparing statement '%s'\n",
		sqlcmd);

	Py_BEGIN_ALLOW_THREADS;
	rc = SQLPrepare(dbcs->hstmt,
			(UCHAR*)sqlcmd,
			sqlcmd_len);
	Py_END_ALLOW_THREADS;

	mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
    }
    
    /* Get number of parameters in statement */
    rc = SQLNumParams(dbcs->hstmt,&sql_inparams);
    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
    if (!params && sql_inparams > 0) {
	PyErr_Format(mxODBC_InterfaceError,
		  "mismatch in number of parameters; expected %i, found none",
		     sql_inparams);
	goto onError;
    }

    DPRINTF("mxODBCursor_Execute: number of params in statement: %i\n",
	    sql_inparams);

    /* Execute with parameters */
    if (params) {

	/* Prepare input variables array */
	vars = cnew(mxODBCursor_Variable,sql_inparams);
	if (!vars) {
	    PyErr_NoMemory();
	    goto onError;
	}
	for (i = 0; i < sql_inparams; i++) {
#ifndef DONT_HAVE_SQLDescribeParam
	    if (dbcs->dbc->have_SQLDescribeParam) {
		rc = SQLDescribeParam(dbcs->hstmt,
				      (UWORD)(i + 1),
				      &vars[i].sqltype,
				      &vars[i].sqllen,
				      &vars[i].sqlscale,
				      &vars[i].nullable);
		mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
		DPRINTF("mxODBCursor_Execute: input column %i: "
			"type=%i len=%i scale=%i nullable=%i\n",
			i,(int)vars[i].sqltype,
			(int)vars[i].sqllen,(int)vars[i].sqlscale,
			(int)vars[i].nullable);
	    }
	    else
#endif
	    {
		/* Setting the sqltype to SQL_TYPE_NULL forces Python
		   type binding for this parameter variable.

		   mxODBCursor_BindInputParameter() will set the
		   var->sql* fields to proper values. 
		*/
		vars[i].sqltype = SQL_TYPE_NULL;
		vars[i].sqllen = 0;
		vars[i].sqlscale = 0;
		vars[i].nullable = 1;
		DPRINTF("mxODBCursor_Execute: input column %i: "
			"binding the variable using Python type\n",
			i);
	    }
	}

	/* Execute one entry at a time

	   XXX Should use arrays and SQLParamOptions() for this... 

	*/
	w = params;
	for (j = 0; j < nrows; j++) {

	    if (PyList_Check(params))
		w = PyList_GET_ITEM(params,j);
	
	    Py_Assert(PyTuple_Check(w),
		      PyExc_TypeError,
		      "parameter list must contain tuples");
		      
	    if (PyTuple_GET_SIZE(w) != sql_inparams) {
		PyErr_Format(mxODBC_InterfaceError,
			     "number of parameters in tuple %i: "
			     "expected %i, found %i",
			     j,sql_inparams,PyTuple_GET_SIZE(w));
		goto onError;
	    }

	    DPRINTF("mxODBCursor_Execute: preparing data for input row %i\n",
		    j);

	    /* Bind parameters for row j */
	    for (i = 0; i < sql_inparams; i++)
		if (mxODBCursor_BindInputParameter(dbcs,
						   &vars[i],
						   PyTuple_GET_ITEM(w,i),
						   i,j))
		    goto onError;

	    DPRINTF("mxODBCursor_Execute: executing cmd for input row %i\n",
		    j);
#if 0
	    /* Close a possibly open statement cursor */
	    rc = SQLFreeStmt(dbcs->hstmt,SQL_CLOSE);
	    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
#endif
	    /* Execute the statement for this entry */

	    Py_BEGIN_ALLOW_THREADS;
	    rc = SQLExecute(dbcs->hstmt);
	    Py_END_ALLOW_THREADS;

	    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

	}
	    
	/* Unbind and free parameters */
	DPRINTF("mxODBCursor_Execute: "
		"reset statement parameters and free buffers\n");
	rc = SQLFreeStmt(dbcs->hstmt,SQL_RESET_PARAMS);
	mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

	for (i = 0; i < sql_inparams; i++)
	    if (vars[i].free_data && vars[i].data) {
		free(vars[i].data);
		vars[i].data = 0;
	    }
	free(vars);
	vars = 0;
    }

    /* Execute without parameters */
    else {
	DPRINTF("mxODBCursor_Execute: executing command without parameters\n");

	/* Execute the statement */

	Py_BEGIN_ALLOW_THREADS;
	rc = SQLExecute(dbcs->hstmt);
	Py_END_ALLOW_THREADS;

	mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);
    }

    /* Prepare output */
    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return 0;

 onError:
    /* Unbind and free parameters; errors produced by this are
       ignored. */
    SQLFreeStmt(dbcs->hstmt,SQL_RESET_PARAMS);
    if (vars) {
	for (i = 0; i < sql_inparams; i++)
	    if (vars[i].free_data && vars[i].data) {
		free(vars[i].data);
		vars[i].data = 0;
	    }
	free(vars);
    }
    return -1;
}

/* --- Methods -------------------------------------------------------- */

/* Note:

   All methods must check whether the cursor and the connection
   (dbcs->dbc) are still open before doing any operation. Some
   Databases or Managers simply dump core in case the connection is
   closed, but the cursor still active.  */

#define dbcs ((mxODBCursorObject*)self)

Py_C_Function( mxODBCursor_close,
	       "close()\n"
	       "Explicitly close a cursor, flushing any remaining fetch data.")
{
    Py_NoArgsCheck();
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");
    if (mxODBCursor_Close(dbcs))
	goto onError;
    Py_ReturnNone();
 onError:
    return NULL;
}

Py_C_Function( mxODBCursor_execute,
	       "execute(sql_string[,params])\n"
	       "Execute a SQL command using the input values from params\n"
	       "if applicable.\n"
	       "- params can be a list of tuples or one single tuple"
	       )
{
    PyObject *sqlcmd;
    PyObject *params = 0;

    Py_Get2Args("O|O",sqlcmd,params);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    Py_Assert(PyString_Check(sqlcmd),
	      PyExc_TypeError,
	      "first argument must be a SQL string");

    /* Execute, possibly reusing the previous command */
#ifndef DISABLE_EXECUTE_CACHE
    if (dbcs->sqlcmd && \
	(dbcs->sqlcmd == sqlcmd || \
	 strcmp(PyString_AS_STRING(dbcs->sqlcmd),
		PyString_AS_STRING(sqlcmd)) == 0)) {
	if (mxODBCursor_Execute(dbcs,
				NULL,
				0,
				params))
	    goto onError;
    }
    else 
#endif
	{
	if (mxODBCursor_Execute(dbcs,
				PyString_AS_STRING(sqlcmd),
				PyString_GET_SIZE(sqlcmd),
				params))
	    goto onError;

	/* Retain a reference to the command string */
	Py_XDECREF(dbcs->sqlcmd);
	Py_INCREF(sqlcmd);
	dbcs->sqlcmd = sqlcmd;
    }

    /* Figure out the return value and return */
    if (dbcs->colcount == 0) {
	if (dbcs->rowcount == 0)
	    /* must be a DDL like CREATE */
	    return PyInt_FromLong(1L);
	else
	    /* must be a DML like UPDATE */
	    return PyInt_FromLong(dbcs->rowcount);
    }
    else
	/* must be a DQL like SELECT */
	Py_ReturnNone();

 onError:
    return NULL;
}

Py_C_Function( mxODBCursor_fetchone,
	       "fetchone()\n"
	       "Return the next query row as tuple; None if no more rows\n"
	       "are available.")
{
    Py_NoArgsCheck();
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    return mxODBCursor_Fetch(dbcs);
 onError:
    return NULL;
}

Py_C_Function( mxODBCursor_fetchmany,
	       "fetchmany([size])\n"
	       "Return size query rows (defaults to cursor.arraysize)\n"
	       "as list of tuples.")
{
    int nrows = 0;

    Py_GetArg("|i",nrows);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (nrows == 0)
	nrows = dbcs->arraysize;
    return mxODBCursor_FetchMany(dbcs,nrows);

 onError:
    return NULL;
}

Py_C_Function( mxODBCursor_fetchall,
	       "fetchall()\n"
	       "Return a list of tuples, one for each row.")
{
    Py_NoArgsCheck();
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    return mxODBCursor_FetchAll(dbcs);
 onError:
    return NULL;
}

/* Catalog functions */

#ifndef DONT_HAVE_SQLTables
Py_C_Function( mxODBCursor_tables,
	       "tables(qualifier,owner,table,type)\n"
	       "Find tables in the data source. See the ODBC API docs\n"
	       "on SQLTable() for details.\n"
	       "- does an implicit cursor.execute()\n"
	       "- returns the rowcount")
{
    PyObject *v[4];
    UCHAR *s[4];
    SWORD l[4];
    int i;
    RETCODE rc;
    
    Py_Get4Args("OOOO",v[0],v[1],v[2],v[3]);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (mxODBCursor_Reset(dbcs))
	goto onError;

    for (i = 0; i < 4; i++)
	if (v[i] == Py_None) {
	    s[i] = NULL;
	    l[i] = 0;
	} else {
	    Py_Assert(PyString_Check(v[i]),
		      PyExc_TypeError,
		      "all arguments must be strings or None");
	    s[i] = (UCHAR*)PyString_AS_STRING(v[i]);
	    l[i] = PyString_GET_SIZE(v[i]);
	}

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLTables(dbcs->hstmt,
		   s[0],l[0],
		   s[1],l[1],
		   s[2],l[2],
		   s[3],l[3]);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return PyInt_FromLong(dbcs->rowcount);
 onError:
    return NULL;
}
#endif

#ifndef DONT_HAVE_SQLTablePrivileges
Py_C_Function( mxODBCursor_tableprivileges,
	       "tableprivileges(qualifier,owner,table)\n")
{
    PyObject *v[3];
    UCHAR *s[3];
    SWORD l[3];
    int i;
    RETCODE rc;
    
    Py_Get3Args("OOO",v[0],v[1],v[2]);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (mxODBCursor_Reset(dbcs))
	goto onError;

    for (i = 0; i < 3; i++)
	if (v[i] == Py_None) {
	    s[i] = NULL;
	    l[i] = 0;
	} else {
	    Py_Assert(PyString_Check(v[i]),
		      PyExc_TypeError,
		      "all arguments must be strings or None");
	    s[i] = (UCHAR*)PyString_AS_STRING(v[i]);
	    l[i] = PyString_GET_SIZE(v[i]);
	}

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLTablePrivileges(dbcs->hstmt,
			    s[0],l[0],
			    s[1],l[1],
			    s[2],l[2]);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return PyInt_FromLong(dbcs->rowcount);
 onError:
    return NULL;
}
#endif

#ifndef DONT_HAVE_SQLColumns
Py_C_Function( mxODBCursor_columns,
	       "columns(qualifier,owner,table,column)\n")
{
    PyObject *v[4];
    UCHAR *s[4];
    SWORD l[4];
    int i;
    RETCODE rc;
    
    Py_Get4Args("OOOO",v[0],v[1],v[2],v[3]);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (mxODBCursor_Reset(dbcs))
	goto onError;

    for (i = 0; i < 4; i++)
	if (v[i] == Py_None) {
	    s[i] = NULL;
	    l[i] = 0;
	} else {
	    Py_Assert(PyString_Check(v[i]),
		      PyExc_TypeError,
		      "all arguments must be strings or None");
	    s[i] = (UCHAR*)PyString_AS_STRING(v[i]);
	    l[i] = PyString_GET_SIZE(v[i]);
	}

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLColumns(dbcs->hstmt,
		    s[0],l[0],
		    s[1],l[1],
		    s[2],l[2],
		    s[3],l[3]);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return PyInt_FromLong(dbcs->rowcount);
 onError:
    return NULL;
}
#endif

#ifndef DONT_HAVE_SQLColumnPrivileges
Py_C_Function( mxODBCursor_columnprivileges,
	       "columnprivileges(qualifier,owner,table,column)\n")
{
    PyObject *v[4];
    UCHAR *s[4];
    SWORD l[4];
    int i;
    RETCODE rc;
    
    Py_Get4Args("OOOO",v[0],v[1],v[2],v[3]);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (mxODBCursor_Reset(dbcs))
	goto onError;

    for (i = 0; i < 4; i++)
	if (v[i] == Py_None) {
	    s[i] = NULL;
	    l[i] = 0;
	} else {
	    Py_Assert(PyString_Check(v[i]),
		      PyExc_TypeError,
		      "all arguments must be strings or None");
	    s[i] = (UCHAR*)PyString_AS_STRING(v[i]);
	    l[i] = PyString_GET_SIZE(v[i]);
	}

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLColumnPrivileges(dbcs->hstmt,
			     s[0],l[0],
			     s[1],l[1],
			     s[2],l[2],
			     s[3],l[3]);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return PyInt_FromLong(dbcs->rowcount);
 onError:
    return NULL;
}
#endif

#ifndef DONT_HAVE_SQLForeignKeys
Py_C_Function( mxODBCursor_foreignkeys,
	       "foreignkeys(primary_qualifier,primary_owner,pimary_table,\n"
	       "            foreign_qualifier,foreign_owner,foreign_table)\n")
{
    PyObject *v[6];
    UCHAR *s[6];
    SWORD l[6];
    int i;
    RETCODE rc;
    
    Py_Get6Args("OOOOOO",v[0],v[1],v[2],v[3],v[4],v[5]);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (mxODBCursor_Reset(dbcs))
	goto onError;

    for (i = 0; i < 6; i++)
	if (v[i] == Py_None) {
	    s[i] = NULL;
	    l[i] = 0;
	} else {
	    Py_Assert(PyString_Check(v[i]),
		      PyExc_TypeError,
		      "all arguments must be strings or None");
	    s[i] = (UCHAR*)PyString_AS_STRING(v[i]);
	    l[i] = PyString_GET_SIZE(v[i]);
	}

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLForeignKeys(dbcs->hstmt,
			s[0],l[0],
			s[1],l[1],
			s[2],l[2],
			s[3],l[3],
			s[4],l[4],
			s[5],l[5]);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return PyInt_FromLong(dbcs->rowcount);
 onError:
    return NULL;
}
#endif

#ifndef DONT_HAVE_SQLPrimaryKeys
Py_C_Function( mxODBCursor_primarykeys,
	       "primarykeys(qualifier,owner,table)\n")
{
    PyObject *v[3];
    UCHAR *s[3];
    SWORD l[3];
    int i;
    RETCODE rc;
    
    Py_Get3Args("OOO",v[0],v[1],v[2]);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (mxODBCursor_Reset(dbcs))
	goto onError;

    for (i = 0; i < 3; i++)
	if (v[i] == Py_None) {
	    s[i] = NULL;
	    l[i] = 0;
	} else {
	    Py_Assert(PyString_Check(v[i]),
		      PyExc_TypeError,
		      "all arguments must be strings or None");
	    s[i] = (UCHAR*)PyString_AS_STRING(v[i]);
	    l[i] = PyString_GET_SIZE(v[i]);
	}

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLPrimaryKeys(dbcs->hstmt,
			s[0],l[0],
			s[1],l[1],
			s[2],l[2]);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return PyInt_FromLong(dbcs->rowcount);
 onError:
    return NULL;
}
#endif

#ifndef DONT_HAVE_SQLProcedures
Py_C_Function( mxODBCursor_procedures,
	       "procedures(qualifier,owner,procedure)\n")
{
    PyObject *v[3];
    UCHAR *s[3];
    SWORD l[3];
    int i;
    RETCODE rc;
    
    Py_Get3Args("OOO",v[0],v[1],v[2]);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (mxODBCursor_Reset(dbcs))
	goto onError;

    for (i = 0; i < 3; i++)
	if (v[i] == Py_None) {
	    s[i] = NULL;
	    l[i] = 0;
	} else {
	    Py_Assert(PyString_Check(v[i]),
		      PyExc_TypeError,
		      "all arguments must be strings or None");
	    s[i] = (UCHAR*)PyString_AS_STRING(v[i]);
	    l[i] = PyString_GET_SIZE(v[i]);
	}

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLProcedures(dbcs->hstmt,
		       s[0],l[0],
		       s[1],l[1],
		       s[2],l[2]);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return PyInt_FromLong(dbcs->rowcount);
 onError:
    return NULL;
}

#endif

#ifndef DONT_HAVE_SQLProcedureColumns
Py_C_Function( mxODBCursor_procedurecolumns,
	       "procedurecolumns(qualifier,owner,procedure,column)\n")
{
    PyObject *v[4];
    UCHAR *s[4];
    SWORD l[4];
    int i;
    RETCODE rc;
    
    Py_Get4Args("OOOO",v[0],v[1],v[2],v[3]);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (mxODBCursor_Reset(dbcs))
	goto onError;

    for (i = 0; i < 4; i++)
	if (v[i] == Py_None) {
	    s[i] = NULL;
	    l[i] = 0;
	} else {
	    Py_Assert(PyString_Check(v[i]),
		      PyExc_TypeError,
		      "all arguments must be strings or None");
	    s[i] = (UCHAR*)PyString_AS_STRING(v[i]);
	    l[i] = PyString_GET_SIZE(v[i]);
	}

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLProcedureColumns(dbcs->hstmt,
			     s[0],l[0],
			     s[1],l[1],
			     s[2],l[2],
			     s[3],l[3]);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return PyInt_FromLong(dbcs->rowcount);
 onError:
    return NULL;
}
#endif

#ifndef DONT_HAVE_SQLSpecialColumns
Py_C_Function( mxODBCursor_specialcolumns,
	       "specialcolumns(qualifier,owner,table,coltype,scope,nullable)\n"
	       "- note: coltype is in a different position than it is in\n"
	       "        the corresponding C call")
{
    PyObject *v[3];
    UCHAR *s[3];
    SWORD l[3];
    int i;
    int coltype,scope, nullable;
    RETCODE rc;
    
    Py_Get6Args("OOOiii",v[0],v[1],v[2],coltype,scope,nullable);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (mxODBCursor_Reset(dbcs))
	goto onError;

    for (i = 0; i < 3; i++)
	if (v[i] == Py_None) {
	    s[i] = NULL;
	    l[i] = 0;
	} else {
	    Py_Assert(PyString_Check(v[i]),
		      PyExc_TypeError,
		      "the first three arguments must be strings or None");
	    s[i] = (UCHAR*)PyString_AS_STRING(v[i]);
	    l[i] = PyString_GET_SIZE(v[i]);
	}

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLSpecialColumns(dbcs->hstmt,
			   (UWORD)coltype,
			   s[0],l[0],
			   s[1],l[1],
			   s[2],l[2],
			   (UWORD)scope,
			   (UWORD)nullable);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return PyInt_FromLong(dbcs->rowcount);
 onError:
    return NULL;
}
#endif

#ifndef DONT_HAVE_SQLStatistics
Py_C_Function( mxODBCursor_statistics,
	       "statistics(qualifier,owner,table,unique,accuracy)\n")
{
    PyObject *v[3];
    UCHAR *s[3];
    SWORD l[3];
    int i;
    int unique, accuracy;
    RETCODE rc;
    
    Py_Get5Args("OOOii",v[0],v[1],v[2],unique,accuracy);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (mxODBCursor_Reset(dbcs))
	goto onError;

    for (i = 0; i < 3; i++)
	if (v[i] == Py_None) {
	    s[i] = NULL;
	    l[i] = 0;
	} else {
	    Py_Assert(PyString_Check(v[i]),
		      PyExc_TypeError,
		      "the first three arguments must be strings or None");
	    s[i] = (UCHAR*)PyString_AS_STRING(v[i]);
	    l[i] = PyString_GET_SIZE(v[i]);
	}

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLStatistics(dbcs->hstmt,
		       s[0],l[0],
		       s[1],l[1],
		       s[2],l[2],
		       (UWORD)unique,
		       (UWORD)accuracy);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return PyInt_FromLong(dbcs->rowcount);
 onError:
    return NULL;
}
#endif

#ifndef DONT_HAVE_SQLGetTypeInfo
Py_C_Function( mxODBCursor_gettypeinfo,
	       "gettypeinfo(sqltypecode)\n")
{
    int sqltypecode;
    RETCODE rc;
    
    Py_GetArg("i",sqltypecode);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    if (mxODBCursor_Reset(dbcs))
	goto onError;

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLGetTypeInfo(dbcs->hstmt,
			(SWORD)sqltypecode);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    if (mxODBCursor_PrepareOutput(dbcs))
	goto onError;

    return PyInt_FromLong(dbcs->rowcount);
 onError:
    return NULL;
}
#endif

/* Dummy functions compliant with DB API 1.0 */

Py_C_Function( mxODBCursor_setinputsizes,
	       "setinputsizes(size)\n"
	       "Unimplemented")
{
    PyObject *sizes;
    
    Py_Get2Args("O!",PyTuple_Type,sizes);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    Py_ReturnNone();
 onError:
    return NULL;
}

Py_C_Function( mxODBCursor_setoutputsizes,
	       "setoutputsizes(size[,col])\n"
	       "Unimplemented")
{
    PyObject *sizes;
    int col;
    
    Py_Get3Args("O!|i",PyTuple_Type,sizes,col);
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    Py_ReturnNone();
 onError:
    return NULL;
}

#ifndef DONT_HAVE_MULTIPLE_RESULT_SETS
/* Dummy functions compliant with DB API 2.0;

   Should check if datasource supports multiple result sets via
   SQLGetInfo() and SQL_MULT_RESULT_SETS... much like for Rollback.

 */

Py_C_Function( mxODBCursor_nextset,
	       "nextset()\n"
	       "Unimplemented")
{
    Py_NoArgsCheck();
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    Py_ReturnNone();
 onError:
    return NULL;
}
#endif

Py_C_Function( mxODBCursor_setcursorname,
	       "setcursorname(name)\n"
	       "Sets the name of the cursor.")
{
    UCHAR *name;
    RETCODE rc;
    
    Py_GetArg("s",name);

    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");
    
    rc = SQLSetCursorName(dbcs->hstmt,
			  (UCHAR *)name,
			  (SWORD)strlen((const char *)name));
    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    Py_ReturnNone();
 onError:
    return NULL;
}

Py_C_Function( mxODBCursor_getcursorname,
	       "getcursorname()\n"
	       "Returns the cursor's name.")
{
    RETCODE rc;
    UCHAR name[100];
    SWORD len;
    
    Py_NoArgsCheck();
    Py_Assert(!dbcs->closed && !dbcs->dbc->closed,
	      mxODBC_ProgError,
	      "cursor or connection already closed");

    rc = SQLGetCursorName(dbcs->hstmt,
			  name,
			  sizeof(name),
			  &len);
    mxODBC_SQLCheck(henv,dbcs->dbc->hdbc,dbcs->hstmt);

    return PyString_FromString((const char *)name);

 onError:
    return NULL;
}

/* --- Slots ----------------------------------------------------------- */

static
int mxODBCursor_Print(mxODBCursorObject *self,
		    FILE *fp, 
		    int flags)
{
    if (self == NULL) return -1;
    fprintf(fp,"<"MXODBC_INTERFACENAME" Cursor object at %lx>",(long)self);
    return 0;
}

static
PyObject *mxODBCursor_GetAttr(mxODBCursorObject *self,
			     char *name)
{
    if (Py_WantAttr(name,"description"))
	return mxODBCursor_Description(self);

    else if (Py_WantAttr(name,"rowcount")) {
	/* Try to update the rowcount if it was undeterminable until
           now (its value being -1 in that case). */
	if (self->rowcount == -1) {
	    RETCODE rc;
	    SDWORD rowcount;

	    rc = SQLRowCount(dbcs->hstmt,&rowcount);
	    /* We ignore any errors this might cause, since querying
	       attributes should not raise any exceptions. */
	    if (rc != SQL_SUCCESS) {
		DPRINTF("mxODBCursor_GetAttr: "
			"SQLRowCount returned an error; ignored.\n");
	    }
	    else
		self->rowcount = rowcount;
	}
	return PyInt_FromLong(self->rowcount);
    }

    else if (Py_WantAttr(name,"colcount"))
	return PyInt_FromLong(self->colcount);

    else if (Py_WantAttr(name,"arraysize"))
	return PyInt_FromLong(self->arraysize);

    else if (Py_WantAttr(name,"datetimeformat"))
	return PyInt_FromLong(self->datetimeformat);

    else if (Py_WantAttr(name,"__members__"))
	return Py_BuildValue("[sssss]",
			     "description","rowcount","colcount",
			     "arraysize","datetimeformat");

    return Py_FindMethod(mxODBCursor_Methods,
			 (PyObject *)self,name);
}

static int
mxODBCursor_SetAttr(mxODBCursorObject *self,
		   char *name,
		   PyObject *v)
{
    if (Py_WantAttr(name,"arraysize")) {
	Py_Assert(PyInt_Check(v),
		  PyExc_TypeError,
		  "arraysize must be an integer");
	self->arraysize = PyInt_AS_LONG(v);
	return 0;
    }

    else if (Py_WantAttr(name,"datetimeformat")) {
	int datetime;
	Py_Assert(PyInt_Check(v),
		  PyExc_TypeError,
		  "datetimeformat must be an integer");
	datetime = PyInt_AS_LONG(v);
	if (datetime < 0 || datetime > MXODBC_MAX_DATETIME)
	    Py_Error(PyExc_TypeError,
		     "datetimeformat out of range");
	self->datetimeformat = datetime;
	return 0;
    }

    Py_ErrorWithArg(PyExc_TypeError,
		    "attribute '%s' unkown or read-only",
		    name);
 onError:
    return -1;
}

/* Python Type Table */

PyTypeObject mxODBCursor_Type = {
    PyObject_HEAD_INIT(0)		/* init at startup ! */
    0,			  		/*ob_size*/
    "mxODBC Cursor",	  		/*tp_name*/
    sizeof(mxODBCursorObject),      	/*tp_basicsize*/
    0,			  		/*tp_itemsize*/
    /* slots */
    (destructor)mxODBCursor_Free,		/*tp_dealloc*/
    (printfunc)mxODBCursor_Print,	  	/*tp_print*/
    (getattrfunc)mxODBCursor_GetAttr,  	/*tp_getattr*/
    (setattrfunc)mxODBCursor_SetAttr,  	/*tp_setattr*/
    (cmpfunc)0,		  		/*tp_compare*/
    (reprfunc)0,		  	/*tp_repr*/
    0,			  		/*tp_as_number*/
    0,					/*tp_as_number*/
    0,					/*tp_as_mapping*/
    (hashfunc)0,			/*tp_hash*/
    (ternaryfunc)0,			/*tp_call*/
    (reprfunc)0,			/*tp_str*/
    (getattrofunc)0, 			/*tp_getattro*/
    (setattrofunc)0, 			/*tp_setattro*/
    0,					/*tp_as_buffer*/
    0,					/*tp_xxx4*/
    (char*) 0				/*tp_doc*/
};

/* Python Method Table */

statichere PyMethodDef mxODBCursor_Methods[] =
{   
    /* DB API interface */
    Py_MethodListEntryNoArgs("fetchone",mxODBCursor_fetchone),
    Py_MethodListEntry("fetchmany",mxODBCursor_fetchmany),
    Py_MethodListEntryNoArgs("fetchall",mxODBCursor_fetchall),
    Py_MethodListEntry("execute",mxODBCursor_execute),
    Py_MethodListEntryNoArgs("close",mxODBCursor_close),
    /* Extensions */
    Py_MethodListEntry("setcursorname",mxODBCursor_setcursorname),
    Py_MethodListEntryNoArgs("getcursorname",mxODBCursor_getcursorname),
    /* Catalog functions */
#ifndef DONT_HAVE_SQLTables
    Py_MethodListEntry("tables",mxODBCursor_tables),
#endif
#ifndef DONT_HAVE_SQLTablePrivileges
    Py_MethodListEntry("tableprivileges",mxODBCursor_tableprivileges),
#endif
#ifndef DONT_HAVE_SQLColumns
    Py_MethodListEntry("columns",mxODBCursor_columns),
#endif
#ifndef DONT_HAVE_SQLColumnPrivileges
    Py_MethodListEntry("columnprivileges",mxODBCursor_columnprivileges),
#endif
#ifndef DONT_HAVE_SQLForeignKeys
    Py_MethodListEntry("foreignkeys",mxODBCursor_foreignkeys),
#endif
#ifndef DONT_HAVE_SQLPrimaryKeys
    Py_MethodListEntry("primarykeys",mxODBCursor_primarykeys),
#endif
#ifndef DONT_HAVE_SQLProcedures
    Py_MethodListEntry("procedures",mxODBCursor_procedures),
#endif
#ifndef DONT_HAVE_SQLProcedureColumns
    Py_MethodListEntry("procedurecolumns",mxODBCursor_procedurecolumns),
#endif
#ifndef DONT_HAVE_SQLSpecialColumns
    Py_MethodListEntry("specialcolumns",mxODBCursor_specialcolumns),
#endif
#ifndef DONT_HAVE_SQLStatistics
    Py_MethodListEntry("statistics",mxODBCursor_statistics),
#endif
#ifndef DONT_HAVE_SQLGetTypeInfo
    Py_MethodListEntry("gettypeinfo",mxODBCursor_gettypeinfo),
#endif
    /* Dummy methods */
    Py_MethodListEntry("setinputsizes",mxODBCursor_setinputsizes),
    Py_MethodListEntry("setoutputsizes",mxODBCursor_setoutputsizes),
#ifndef DONT_HAVE_MULTIPLE_RESULT_SETS
    Py_MethodListEntryNoArgs("nextset",mxODBCursor_nextset),
#endif
    {NULL,NULL} /* end of list */
};

#undef dbcs

/* --- Database Connection Object --------------------------------------*/

staticforward PyMethodDef mxODBC_Methods[];

/* --- allocation --- */

static
int mxODBC_InitConnection(mxODBCObject *dbc,
			  int clearAC)
{
    RETCODE rc;
    UWORD check;
    
    dbc->datetimeformat = MXODBC_DATETIME_DATETIME;
    dbc->closed = 0;

#ifdef USE_SQLTYPE_BINDING
    dbc->bindmethod = MXODBC_BIND_USING_SQLTYPE;
#else
    dbc->bindmethod = MXODBC_BIND_USING_PYTHONTYPE;
#endif

    /* Check some SQL API we use */
#ifdef DONT_HAVE_SQLDescribeParam
    dbc->have_SQLDescribeParam = 0;
#else
# ifdef HAVE_SQLGetFunctions
    check = 0;
    rc = SQLGetFunctions(dbc->hdbc,SQL_API_SQLDESCRIBEPARAM,&check);
    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
    dbc->have_SQLDescribeParam = check;
    /* Fallback solution... */
    if (!check)
	dbc->bindmethod = MXODBC_BIND_USING_PYTHONTYPE;
# else
    dbc->have_SQLDescribeParam = 1;
# endif
#endif

    /* Some DBs don't like this, so we make it optional... */
    if (clearAC) {
	rc = SQLSetConnectOption(dbc->hdbc, SQL_AUTOCOMMIT, 0);
	mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
    }

    return 0;
    
 onError:
    check = 0; /* To shut up -Wall... */
    return -1;
}

static
PyObject *mxODBC_New(char *server,
		     char *uid,
		     char *passwd,
		     int clearAC)
{
    mxODBCObject *dbc;
    RETCODE rc;
    int allocated = 0;
    int connected = 0;

    DPRINTF("mxODBC_New: server='%s', uid='%s', passwd='%s', clearAC=%i\n",
	    server,uid,passwd,clearAC);

    dbc = PyObject_NEW(mxODBCObject,&mxODBC_Type);
    if (dbc == NULL)
	return NULL;

    rc = SQLAllocConnect(henv,&dbc->hdbc);
    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
    allocated = 1;

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLConnect(dbc->hdbc,
		    (UCHAR*)server, SQL_NTS, 
		    (UCHAR*)uid, SQL_NTS,
		    (UCHAR*)passwd, SQL_NTS);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLErrorCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
    connected = 1;

    /* Init. vars */
    if (mxODBC_InitConnection(dbc,clearAC))
	goto onError;

    DPRINTF("mxODBC_New: created new connection at 0x%lx\n",(long)dbc);
    return (PyObject *)dbc;

 onError:
    if (connected)
	SQLDisconnect(dbc->hdbc);
    if (allocated)
	SQLFreeConnect(dbc->hdbc);
    /* We don't call the deallocator to avoid confusion... */
    PyMem_DEL(dbc);
    return NULL;
}

#ifdef HAVE_SQLDriverConnect
static
PyObject *mxODBC_New_UseDriverConnect(char *dsn,
				      int clearAC)
{
    mxODBCObject *dbc;
    RETCODE rc;
    UCHAR dsn_out[512];
    SWORD dsn_out_len;
    int allocated = 0;
    int connected = 0;
    
    DPRINTF("mxODBC_New_UseDriverConnect: dsn='%s', clearAC=%i\n",
	    dsn,clearAC);

    dbc = PyObject_NEW(mxODBCObject,&mxODBC_Type);
    if (dbc == NULL)
	return NULL;

    rc = SQLAllocConnect(henv,&dbc->hdbc);
    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
    allocated = 1;

    Py_BEGIN_ALLOW_THREADS;
    rc = SQLDriverConnect(dbc->hdbc,
			  (HWND)NULL,
			  (UCHAR *)dsn, SQL_NTS,
			  dsn_out,sizeof(dsn_out),&dsn_out_len,
			  SQL_DRIVER_NOPROMPT);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
    connected = 1;

    /* Init. vars */
    if (mxODBC_InitConnection(dbc,clearAC))
	goto onError;

    DPRINTF("mxODBC_New_UseDriverConnect: created new connection at 0x%lx\n",
	    (long)dbc);
    return (PyObject *)dbc;

 onError:
    if (connected)
	SQLDisconnect(dbc->hdbc);
    if (allocated)
	SQLFreeConnect(dbc->hdbc);
    /* We don't call the deallocator to avoid confusion... */
    PyMem_DEL(dbc);
    return NULL;
}
#endif

/* --- deallocation --- */

/* Close the connection. Returns 0 on success. 

   If report_errors is 1, exceptions will be raised and reported back
   to the caller with -1 return value. With report_errors set to 0,
   exceptions are not reported and 0 is always returned.

*/

static
int mxODBC_Close(mxODBCObject *dbc,
		 int report_errors)
{
    RETCODE rc;
    DPRINTF("mxODBC_Close: called for connection at 0x%lx, closed=%i\n",
	    (long)dbc,dbc->closed);

    if (dbc->closed)
	return 0;

    /* We don't want to accidentaly clear an error that wasn't set by
       us... */
    if (PyErr_Occurred())
	report_errors = -1;
    
    DPRINTF("mxODBC_Close:  disconnect\n");
    rc = SQLDisconnect(dbc->hdbc);
#ifndef DONT_HAVE_TRANSACTIONS
    if (rc == SQL_ERROR) {
	DPRINTF("mxODBC_Close:  disconnect failed; "
		"rolling back the transaction\n");
	if (mxODBC_Rollback(dbc) && report_errors)
	    goto onError;
	DPRINTF("mxODBC_Close:  disconnect\n");
	rc = SQLDisconnect(dbc->hdbc);
    }
#endif
    mxODBC_SQLErrorCheck_Only(henv,dbc->hdbc,SQL_NULL_HSTMT);

    DPRINTF("mxODBC_Close:  free connection\n");
    rc = SQLFreeConnect(dbc->hdbc);
    mxODBC_SQLErrorCheck_Only(henv,dbc->hdbc,SQL_NULL_HSTMT);

    dbc->hdbc = 0;
    dbc->closed = 1;

    /* Now check for any errors that might have to be reported to
       the caller */
    if (PyErr_Occurred()) {
	DPRINTF("mxODBC_Close:  errors occurred\n");
	if (report_errors)
	    goto onError;
	else if (report_errors > 0)
	    PyErr_Clear();
    }
    return 0;

 onError:
    return -1;
}

static
void mxODBC_Free(mxODBCObject *dbc)
{
    DPRINTF("mxODBC_Free: called for connection at 0x%lx\n",(long)dbc);
    mxODBC_Close(dbc,0);
    PyMem_DEL(dbc);
}

/* --- Methods -------------------------------------------------------- */

#define dbc ((mxODBCObject*)self)

Py_C_Function( mxODBC_cursor,
	       "cursor([name])\n"
	       "Return a new cursor object for this connection.\n"
	       "A name may be given to the cursor; otherwise the driver\n"
	       "will create a generated name for it.")
{
    char *name = NULL;
    
    Py_GetArg("|s",name);

    return mxODBCursor_New((mxODBCObject*)self,name);

 onError:
    return NULL;
}

Py_C_Function( mxODBC_close,
	       "close()\n"
	       "Explicitly close a connection.")
{
    Py_NoArgsCheck();
    if (mxODBC_Close(dbc,1))
	goto onError;
    Py_ReturnNone();
 onError:
    return NULL;
}

Py_C_Function( mxODBC_commit,
	       "commit()\n"
	       "Commit any work pending.")
{
    RETCODE rc;

    Py_NoArgsCheck();
    Py_Assert(!dbc->closed,
	      mxODBC_ProgError,
	      "connection already closed");
    
    DPRINTF("mxODBC_commit: Committing changes for connection at 0x%lx\n",
	    (long)dbc);

#ifndef DONT_HAVE_TRANSACTIONS
    Py_BEGIN_ALLOW_THREADS;
    rc = SQLTransact(henv,dbc->hdbc,SQL_COMMIT);
    Py_END_ALLOW_THREADS;

    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
#else
    rc = SQL_SUCCESS;
    DPRINTF("mxODBC_commit:  transactions are not supported: ignored\n");
#endif

    Py_ReturnNone();
 onError:
    return NULL;
}

#ifndef DONT_HAVE_TRANSACTIONS
Py_C_Function( mxODBC_rollback,
	       "rollback()\n"
	       "Rollback any work pending.")
{
    Py_NoArgsCheck();

    if (mxODBC_Rollback(dbc))
	goto onError;

    Py_ReturnNone();
 onError:
    return NULL;
}
#endif

Py_C_Function( mxODBC_setconnectoption,
	       "setconnectoption(option,value)\n"
	       "Set a connection option.\n"
	       "NOTE: Only integer options/values are supported !!!\n"
	       "Refer to the ODBC API for more information -- this function\n"
	       "is a direct interface to SQLSetConnectOption")
{
    int option;
    long value;
    char buffer[16]; /* passing UDWORD directly dumps core for ADABAS... */
    RETCODE rc;

    Py_Get2Args("il",option,value);
    Py_Assert(!dbc->closed,
	      mxODBC_ProgError,
	      "connection already closed");
    
    memset(buffer,0,sizeof(buffer));
    memcpy(buffer,(void *)&value,sizeof(UDWORD));
    rc = SQLSetConnectOption(dbc->hdbc,
			     (UWORD)option,
			     *(UDWORD *)buffer);
    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
    Py_ReturnNone();
 onError:
    return NULL;
}

Py_C_Function( mxODBC_getconnectoption,
	       "getconnectoption(option)\n"
	       "Get the current value of a connection option.\n"
	       "NOTE: Only integer options are supported !!!\n"
	       "Refer to the ODBC API for more information -- this function\n"
	       "is a direct interface to SQLGetConnectOption")
{
    int option;
    char value[16]; /* UDWORD dumps core... */
    RETCODE rc;

    Py_GetArg("i",option);
    Py_Assert(!dbc->closed,
	      mxODBC_ProgError,
	      "connection already closed");
    
    rc = SQLGetConnectOption(dbc->hdbc,
			     (UWORD)option,
			     (PTR)value);
    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);
    return PyInt_FromLong(*(long*)value);
 onError:
    return NULL;
}

Py_C_Function( mxODBC_getinfo,
	       "getinfo(id)\n"
	       "Get information about driver capabilities.\n"
	       "Returns a tuple (long, string); decoding the information\n"
	       "is left to the caller. Refer to the ODBC API for more\n"
	       "information -- this function is a direct interface to\n"
	       "SQLGetInfo")
{
    int id;
    char value[1024];
    SWORD value_len;
    RETCODE rc;

    Py_GetArg("i",id);
    Py_Assert(!dbc->closed,
	      mxODBC_ProgError,
	      "connection already closed");
    
    rc = SQLGetInfo(dbc->hdbc,
		    (UWORD)id,
		    (PTR)value,
		    (SWORD)sizeof(value),
		    &value_len);
    mxODBC_SQLCheck(henv,dbc->hdbc,SQL_NULL_HSTMT);

    Py_Return3Args("is#",*(long *)value,value,value_len);

 onError:
    return NULL;
}

/* --- Slots --------------------------------------------------------- */

static
int mxODBC_Print(mxODBCObject *self,
		FILE *fp, 
		int flags)
{
    if (self == NULL) 
	return -1;
    fprintf(fp,"<"MXODBC_INTERFACENAME" Connection object at %lx>",(long)self);
    return 0;
}

static
PyObject *mxODBC_GetAttr(mxODBCObject *self,
			char *name)
{
    if (Py_WantAttr(name,"datetimeformat"))
	return PyInt_FromLong(self->datetimeformat);

    else if (Py_WantAttr(name,"closed"))
	return PyInt_FromLong((long)self->closed);

    else if (Py_WantAttr(name,"bindmethod"))
	return PyInt_FromLong(self->bindmethod);

    else if (Py_WantAttr(name,"__members__"))
	return Py_BuildValue("[sss]",
			     "datetimeformat","closed","bindmethod");

    return Py_FindMethod(mxODBC_Methods, (PyObject *)self, (char *)name);
}

static int
mxODBC_SetAttr(mxODBCObject *self,
	       char *name,
	       PyObject *v)
{
    if (Py_WantAttr(name,"datetimeformat")) {
	int datetime;
	Py_Assert(PyInt_Check(v),
		  PyExc_TypeError,
		  "datetimeformat must be an integer");
	datetime = PyInt_AS_LONG(v);
	if (datetime < 0 || datetime > MXODBC_MAX_DATETIME)
	    Py_Error(PyExc_TypeError,
		     "datetimeformat out of range");
	self->datetimeformat = datetime;
	return 0;
    }

    else if (Py_WantAttr(name,"bindmethod")) {
	int bindmethod;
	Py_Assert(PyInt_Check(v),
		  PyExc_TypeError,
		  "bindmethod must be an integer");
	bindmethod = PyInt_AS_LONG(v);
	if (bindmethod < 1 || bindmethod > MXODBC_MAX_BIND_METHOD)
	    Py_Error(PyExc_TypeError,
		     "bindmethod out of range");
	self->bindmethod = bindmethod;
	return 0;
    }

    PyErr_Format(PyExc_TypeError,
		 "attribute \"%s\" unkown or read-only",
		 name);
 onError:
    return -1;
}

/* Python Type Table */

PyTypeObject mxODBC_Type = {
    PyObject_HEAD_INIT(0) 		/* init at startup ! */
    0,			  		/*ob_size*/
    "mxODBC Connection",	  	/*tp_name*/
    sizeof(mxODBCObject),	  	/*tp_basicsize*/
    0,			  		/*tp_itemsize*/
    /* slots */
    (destructor)mxODBC_Free,		/*tp_dealloc*/
    (printfunc)mxODBC_Print,	  	/*tp_print*/
    (getattrfunc)mxODBC_GetAttr,  	/*tp_getattr*/
    (setattrfunc)mxODBC_SetAttr, 	/*tp_setattr*/
    (cmpfunc)0,		  		/*tp_compare*/
    (reprfunc)0,		 	/*tp_repr*/
    0,			  		/*tp_as_number*/
    0,					/*tp_as_number*/
    0,					/*tp_as_mapping*/
    (hashfunc)0,			/*tp_hash*/
    (ternaryfunc)0,			/*tp_call*/
    (reprfunc)0,			/*tp_str*/
    (getattrofunc)0, 			/*tp_getattro*/
    (setattrofunc)0, 			/*tp_setattro*/
    0,					/*tp_as_buffer*/
    0,					/*tp_xxx4*/
    (char*) 0				/*tp_doc*/
};

/* Python Method Table */

statichere PyMethodDef mxODBC_Methods[] =
{   
    Py_MethodListEntry("cursor",mxODBC_cursor),
    Py_MethodListEntryNoArgs("close",mxODBC_close),
    Py_MethodListEntryNoArgs("commit",mxODBC_commit),
#ifndef DONT_HAVE_TRANSACTIONS
    Py_MethodListEntryNoArgs("rollback",mxODBC_rollback),
#endif
    Py_MethodListEntry("setconnectoption",mxODBC_setconnectoption),
    Py_MethodListEntry("getconnectoption",mxODBC_getconnectoption),
    Py_MethodListEntry("getinfo",mxODBC_getinfo),
    {NULL,NULL} /* end of list */
};

/* --- constructors -------------------------------------------------------- */

Py_C_Function_WithKeywords( 
	       mxODBC_Connect,
	       "Connect(dsn,user='',password=''[,clear_auto_commit])\n\n"
	       "Returns a new connection object.")
{
    char *dsn,*user="",*password="";
    int clear_auto_commit;

#ifdef DONT_CLEAR_AUTOCOMMIT
    clear_auto_commit = 0;
#else
    clear_auto_commit = 1;
#endif
    Py_KeywordsGet4Args("s|ssi",dsn,user,password,clear_auto_commit);
    return mxODBC_New(dsn,user,password,clear_auto_commit);
 onError:
    return NULL;
}

#ifdef HAVE_ODBC_DRIVER_CONNECT
Py_C_Function( mxODBC_DriverConnect,
	       "DriverConnect(dsn_string[,clear_auto_commit])\n\n"
	       "Returns a new connection object.")
{
    char *dsn;
    int clearAC;

#ifdef DONT_CLEAR_AUTOCOMMIT
    clearAC = 0;
#else
    clearAC = 1;
#endif
    Py_Get2Args("s|i",dsn,clearAC);
    return mxODBC_New_UseDriverConnect(dsn,clearAC);
 onError:
    return NULL;
}
#endif

#undef dbc

/* --- module init --------------------------------------------------------- */

/* Python Method Table */

static PyMethodDef Module_methods[] =
{   
    Py_MethodWithKeywordsListEntry("Connect",mxODBC_Connect),
    Py_MethodWithKeywordsListEntry("connect",mxODBC_Connect),
#ifdef HAVE_ODBC_DRIVER_CONNECT
    Py_MethodListEntry("DriverConnect",mxODBC_DriverConnect),
#endif
    /* DB-API compatible old name: */
    Py_MethodListEntry(MXODBC_MODULE,mxODBC_Connect),
    {NULL,NULL} /* end of list */
};


/* Helper for adding integer constants
   - check for errors with PyErr_Occurred()
 */
static 
void insint(PyObject *dict,
	    char *name,
	    int value)
{
    PyObject *v = PyInt_FromLong((long)value);
    PyDict_SetItemString(dict, name, v);
    Py_XDECREF(v);
}

/* Helper for adding integer constants to a dictionary. Check for
   errors with PyErr_Occurred() */
static 
void insstr(PyObject *dict,
	    char *name,
	    char *value)
{
    PyObject *v = PyString_FromString(value);
    PyDict_SetItemString(dict, name, v);
    Py_XDECREF(v);
}

static
void mxODBCModule_Cleanup(void)
{
    RETCODE rc;

    rc = SQLFreeEnv(henv);
    /* Can't put the error anywhere so we'll just ignore it. */
}

/* Create PyMethodObjects and register them in the module's dict */
MX_EXPORT(void)
     initmxODBC(void)
{
    PyObject *module, *moddict, *d, *c;
    int i;
    
    /* Init type objects */
    PyType_Init(mxODBC_Type);
    PyType_Init(mxODBCursor_Type);

    /* create module */
    module = Py_InitModule4(MXODBC_MODULE, /* Module name */
			    Module_methods, /* Method list */
			    Module_docstring, /* Module doc-string */
			    (PyObject *)NULL, /* always pass this as *self */
			    PYTHON_API_VERSION); /* API Version */

    /* Init ODBC API */
    if (SQLAllocEnv(&henv) != SQL_SUCCESS)
	Py_Error(mxODBC_InterfaceError,
		 "can't initialize ODBC API");

    /* Register cleanup function */
    if (Py_AtExit(mxODBCModule_Cleanup))
	/* XXX what to do if we can't register that function ??? */
	i=0;

    /* Load mxDateTime module API */
    if (mxDateTime_ImportModuleAndAPI())
	goto onError;

    /* Add some constants to the module's dict */
    moddict = PyModule_GetDict(module);
    PyDict_SetItemString(moddict, 
			 "__version__",
			 PyString_FromString(VERSION));

    /* Init exceptions */
    if ((mxODBC_Error = insexc(moddict,
			       "Error",PyExc_StandardError)) == NULL)
	goto onError;
    /* alias */
    if (PyDict_SetItemString(moddict, "error", mxODBC_Error) < 0)
	goto onError;
    if ((mxODBC_Warning = insexc(moddict,
				 "Warning",PyExc_StandardError))
	== NULL)
	goto onError;
    if ((mxODBC_InterfaceError = insexc(moddict,
					"InterfaceError",mxODBC_Error))
	== NULL)
	goto onError;
    if ((mxODBC_DatabaseError = insexc(moddict,
				       "DatabaseError",mxODBC_Error))
	== NULL)
	goto onError;
    if ((mxODBC_InternalError = insexc(moddict,
				       "InternalError",mxODBC_DatabaseError))
	== NULL)
	goto onError;
    if ((mxODBC_OpError = insexc(moddict,
				 "OperationalError",mxODBC_DatabaseError))
	== NULL)
	goto onError;
    if ((mxODBC_ProgError = insexc(moddict,
				   "ProgrammingError",mxODBC_DatabaseError))
	== NULL)
	goto onError;
    if ((mxODBC_IntegrityError = insexc(moddict,
					"IntegrityError",mxODBC_DatabaseError))
	== NULL)
	goto onError;
    if ((mxODBC_DataError = insexc(moddict,
				   "DataError",mxODBC_DatabaseError))
	== NULL)
	goto onError;
    if ((mxODBC_NotSuppError = insexc(moddict,
				      "NotSupportedError",mxODBC_DatabaseError))
	== NULL)
	goto onError;

    mxODBC_ErrorDict = PyDict_New();
    if (mxODBC_ErrorDict == NULL)
	goto onError;
    for (i = 0; i < sizeof(mxODBC_ErrorCodeTranslations) /
	            sizeof(mxODBC_ErrorCodeTranslation); i++) {
	PyObject *errorobj = *mxODBC_ErrorCodeTranslations[i].errorobj;
	Py_INCREF(errorobj);
	if (PyDict_SetItemString(mxODBC_ErrorDict,
				 mxODBC_ErrorCodeTranslations[i].sqlstate,
				 errorobj))
	    goto onError;
    }
    if (PyDict_SetItemString(moddict,"errorclass",mxODBC_ErrorDict))
	goto onError;
    
    /* Data types */

    d = PyDict_New();
    if (d == NULL)
	goto onError;
    for (i = 0; i < sizeof(mxODBC_TypeCodeTranslations) /
	            sizeof(mxODBC_TypeCodeTranslation); i++) {
	PyObject *v,*w;
	v = PyString_FromString(mxODBC_TypeCodeTranslations[i].sqlstring);
	if (v == NULL)
	    goto onError;
	w = PyInt_FromLong((long)mxODBC_TypeCodeTranslations[i].sqlcode);
	if (w == NULL)
	    goto onError;
	/* Add to sqltype dict */
	if (PyDict_SetItem(d,
			   w,v))
	    goto onError;
	/* Add to module dict */
	if (PyDict_SetItem(moddict,
			   v,w))
	    goto onError;
    }
    if (PyDict_SetItemString(moddict,"sqltype",d))
	goto onError;
    
    /* SQL Class */

    c = mxODBC_SQL_Class();
    if (c == NULL)
	goto onError;
    if (PyDict_SetItemString(moddict,"SQL",c))
	goto onError;
    
    /* Date/Time format values */
    insint(moddict,"DATETIME_DATETIMEFORMAT",MXODBC_DATETIME_DATETIME);
    insint(moddict,"TIMEVALUE_DATETIMEFORMAT",MXODBC_FLOAT_DATETIME);
    insint(moddict,"TUPLE_DATETIMEFORMAT",MXODBC_TUPLE_DATETIME);
    insint(moddict,"STRING_DATETIMEFORMAT",MXODBC_STRING_DATETIME);

    /* Bind method values */
    insint(moddict,"BIND_USING_PYTHONTYPE",MXODBC_BIND_USING_PYTHONTYPE);
    insint(moddict,"BIND_USING_SQLTYPE",MXODBC_BIND_USING_SQLTYPE);
    insint(moddict,"TIMEVALUE_DATETIMEFORMAT",MXODBC_FLOAT_DATETIME);
    insint(moddict,"TUPLE_DATETIMEFORMAT",MXODBC_TUPLE_DATETIME);
    insint(moddict,"STRING_DATETIMEFORMAT",MXODBC_STRING_DATETIME);

    /* DB API 2.0 globals */
    insint(moddict,"threadlevel",MXODBC_THREADLEVEL);
    insstr(moddict,"apilevel",MXODBC_APILEVEL);
    insstr(moddict,"paramstyle","qmark");

    /* Type objects */
    Py_INCREF(&mxODBC_Type);
    PyDict_SetItemString(moddict, "mxODBCType",
			 (PyObject *)&mxODBC_Type);
    Py_INCREF(&mxODBCursor_Type);
    PyDict_SetItemString(moddict, "mxODBCursorType",
			 (PyObject *)&mxODBCursor_Type);


    DPRINTF("sizeof(mxODBCursor_Variable)=%i\n",sizeof(mxODBCursor_Variable));
    
 onError:
    /* Check for errors and report them */
    if (PyErr_Occurred())
	Py_ReportModuleInitError(MXODBC_MODULE);
    return;
}
